import threading
import time
import fractions
import uuid
import logger
from TestInput import TestInputServer

import mc_bin_client
from membase.api.rest_client import RestConnection
from memcached.helper.data_helper import VBucketAwareMemcached


class FakeMemcachedClient(object):
    def __init__(self, *args):
        self.db = {}
    def set(self, key, exp, flags, val):
        self.db[key] = val
    def get(self, key):
        return (0,0,self.db[key])
    def add(self, key, exp, flags, val):
        self.db[key] = val
    def delete(self, key):
        del self.db[key]
    def sasl_auth_plain(self, *args):
        pass


class LoadThread(threading.Thread):
    def __init__(self, load_info, server_index):
        threading.Thread.__init__(self)
        self.daemon = True

        self.log = logger.Logger()
        # thread state info
        self.stopped = False
        self.paused = True

        self.mutation_index = 0
        self.get_index = 0
        self.value_failures = 0
        self.backoff = 0

        self.mutation_index_max = 0
        self.items = 0
        self.operations = 0
        self.time = 0
        self.size = 0
        self.operation_rate = 0

        threads = int(load_info['operation_info'].get('threads', 0))

        # cache info
        self.cache_data = load_info['operation_info'].get('cache_data', False)
        self.data_cache = {}

        # server info
        self.server_ip = load_info['server_info'][server_index].ip
        self.bucket_name = load_info['memcached_info'].get('bucket_name', '')
        self.bucket_password = load_info['memcached_info'].get('bucket_password', '')

        # operation info
        self.create = load_info['operation_info']['create_percent'] / fractions.gcd(load_info['operation_info']['create_percent'], 100 - load_info['operation_info']['create_percent'])
        self.nocreate = (100 - load_info['operation_info']['create_percent']) / fractions.gcd(load_info['operation_info']['create_percent'], 100 - load_info['operation_info']['create_percent'])
        self.operation_sequence = []
        for op in load_info['operation_info']['operation_distribution']:
            for i in range(load_info['operation_info']['operation_distribution'][op]):
                self.operation_sequence.append(op)

        self.valuesize_sequence = []
        for op in load_info['operation_info']['valuesize_distribution']:
            for i in range(load_info['operation_info']['valuesize_distribution'][op]):
                self.valuesize_sequence.append(op)

        self.max_operation_rate = int(load_info['operation_info'].get('operation_rate', 0) / threads)

        self.uuid = uuid.uuid4()
        self.name = str(self.uuid) + self.name

        # limit info
        # all but time needs to be divided equally amongst threads
        self.limit_items = int(load_info['limit_info'].get('items', 0) / threads)
        self.limit_operations = int(load_info['limit_info'].get('operations', 0) / threads)
        self.limit_time = int(load_info['limit_info'].get('time', 0))
        self.limit_size = int(load_info['limit_info'].get('size', 0) / threads)
        self.poxi = self._poxi()
        print "now connected to {0} memcacheds".format(len(self.poxi.memcacheds))
        # connect
        #self.server should be vbucketaware memcached
#        self.server = mc_bin_client.MemcachedClient(self.server_ip, self.server_port)
#        if self.bucket_name or self.bucket_password:
#            self.server.sasl_auth_plain(self.bucket_name,self.bucket_password)


    def _poxi(self):
        tServer = TestInputServer()
        tServer.ip = self.server_ip
        tServer.rest_username = "Administrator"
        tServer.rest_password = "password"
        tServer.port = 8091
        rest = RestConnection(tServer)
        return VBucketAwareMemcached(rest, self.bucket_name)

    def run(self):
        while True:
            # handle pause/stop
            while self.paused:
                if self.stopped:
                    return
                time.sleep(1)
            if self.stopped:
                return

            start_time = time.time()

            # stop thread if we hit a limit (first limit we hit ends the thread)
            if self.limit_items and self.items > self.limit_items:
                self.log.info("items count limit reached")
                return
            if self.limit_operations and self.operations > self.limit_operations:
                self.log.info("operations limit reached")
                return
            if self.limit_time and self.time > self.limit_time:
                self.log.info("time limit reached")
                return
            if self.limit_size and self.size > self.limit_size:
                self.log.info("size limit reached")
                return

            # rate limit if we need to
            if self.max_operation_rate and self.operation_rate > self.max_operation_rate:
                time.sleep(1.0/float(self.max_operation_rate))

            # do the actual work
            operation = self.get_operation()
            if operation == 'set':
                key = self.name + '_' + `self.get_mutation_key()`
                try:
#                    print `self.mutation_index` + " : " + `self.get_mutation_key()`
                    self.poxi.memcached(key).set(key, 0, 0, self.get_data())
                    self.operations += 1
                    self.backoff -= 1

                    # update number of items
                    # for now the only time we have new items is with ever increasing mutation key indexes
                    if self.get_mutation_key() > self.mutation_index_max:
                        self.mutation_index_max = self.get_mutation_key()
                        # looks like this will miss the first mutation
                        self.items += 1

                    # TODO: verify that this works, we may need to take the second to max index
                    # update the size of all data (values, not keys) that is in the system
                    # this can be either from new keys or overwriting old keys
                    prev_indexes = self.get_mutation_indexes(self.get_mutation_key())
                    prev_size = 0
                    if prev_indexes:
                        prev_size = self.get_data_size(max(prev_indexes))
                    self.size += self.get_data_size() - prev_size

                    self.mutation_index += 1
                except mc_bin_client.MemcachedError as e:
                    self.poxi.done()
                    self.poxi = self._poxi()
                    print "now connected to {0} memcacheds".format(len(self.poxi.memcacheds))
                    if self.backoff < 0:
                        self.backoff = 0
                    if self.backoff > 30:
                        self.backoff = 30
                    self.backoff += 1
                    # temporary error
#                    if e.status == 134:
#                        time.sleep(self.backoff)
#                    else:
#                        print `time.time()` + ' ' + self.name + ' set(' + `self.backoff` + ') ',
#                        print e
#                        time.sleep(self.backoff)
            elif operation == 'get':
                key = self.name + '_' + `self.get_get_key()`
                try:
                    vdata = self.poxi.memcached(key).get(key)
                    self.operations += 1
                    self.backoff -= 1
                    data = vdata[2]
                    try:
                        data_expected = self.get_data(max(self.get_mutation_indexes(self.get_get_key())))
                        if data != data_expected:
                            self.value_failures += 1
                            raise
                    except Exception as e:
                        print e
                        print "create: " + `self.create`
                        print "nocreate: " + `self.nocreate`
                        print "get_index: " + `self.get_index`
                        print "get_key: " + `self.get_get_key()`
                        print "mutation_index_max: " + `self.mutation_index_max`
                        print "mutation_indexes: " + `self.get_mutation_indexes(self.get_get_key())`
                        print "getting data for mutation index: " + `max(self.get_mutation_indexes(self.get_get_key()))`
                        print "got:      \'" + data + "\'"
                        raise ValueError
                    self.get_index += 1
                except mc_bin_client.MemcachedError as e:
                    if self.backoff < 0:
                        self.backoff = 0
                    if self.backoff > 30:
                        self.backoff = 30
                    self.backoff += 1
                    print `time.time()` + ' ' + self.name + ' get(' + `self.backoff` + ') ',
                    print e
                    time.sleep(self.backoff)

            end_time = time.time()
            self.time += (end_time-start_time)
            self.operation_rate = (float(self.operations) / self.time)

    # get the current operation based on the get and mutation indexes
    def get_operation(self):
        return self.operation_sequence[(self.mutation_index + self.get_index) % len(self.operation_sequence)]

    # mutation_index -> mutation_key : based on create/nocreate
    def get_mutation_key(self, index=None):
        if index == None:
            index = self.mutation_index
        return index-self.nocreate*(index/(self.create+self.nocreate))

    # get_index -> get_key : based on get_index % mutation_index
    def get_get_key(self):
        return self.get_index % self.get_mutation_key(self.mutation_index)

    # key -> mutation_indexes
    def get_mutation_indexes(self, key):
        # starting point to find an index
        s=key*(self.nocreate+self.create)/self.create

        mutation_indexes=[]

        # for now we will scan all possible gcs even though we knows the step size
        # if we could guarentee that our calculated s was actually a gc it would be faster
        # scan a range (right now nocreate^2) incrementing by 1
        #  once we find a valid index, increment by nocreate
        index = s-(self.nocreate*self.nocreate)
        if index < 0:
            index = 0
        index_max = s+(self.nocreate*self.nocreate)
        incr = 1
        while index <= index_max:
            if index >= 0 and index <= self.mutation_index and self.get_mutation_key(index) == key:
                incr = self.nocreate
                mutation_indexes.append(index)
            index += incr
        return mutation_indexes

    # mutation_index -> mutation_data : based on create/nocreate
    def get_data(self, index=None):
        if index == None:
            index = self.mutation_index

        valuesize = self.valuesize_sequence[index % len(self.valuesize_sequence)]
        if self.cache_data:
            if not valuesize in self.data_cache:
                self.data_cache[valuesize] = (str(uuid.uuid3(self.uuid,`index`)) * (1+valuesize/36))[:valuesize]
            return `index` + self.data_cache[valuesize]
        else:
            return (str(uuid.uuid3(self.uuid,`index`)) * (1+valuesize/36))[:valuesize]

    # mutation_index -> mutation_data : based on create/nocreate
    # shortcut for getting the expected size of a mutation without generating the data
    def get_data_size(self, index=None):
        if index == None:
            index = self.mutation_index

        valuesize = self.valuesize_sequence[index % len(self.valuesize_sequence)]
        return valuesize


class LoadRunner(object):
    def __init__(self, load_info, dryrun=False):
        if dryrun:
            mc_bin_client.MemcachedClient = FakeMemcachedClient

        self.stopped = False
        self.paused = True

        self.threads = []
        self.num_servers = len(load_info['server_info'])
        self.num_threads = int(load_info['operation_info']['threads'])
        for i in range(self.num_threads):
            t = LoadThread(load_info, i % self.num_servers)
            t.start()
            self.threads.append(t)

    # start running load against server
    # this is run in a seperate thread(s)
    def start(self):
        self.stopped = False
        self.paused = False
        for t in self.threads:
            t.stopped = False
            t.paused = False
        
    # pause load but keep track of where we are
    def pause(self):
        self.paused = True
        for t in self.threads:
            t.paused = True

    # stop load and reset back to start
    def stop(self):
        self.stopped = True
        for t in self.threads:
            t.stopped = True

    # dump entire sequence of operations to a file
    def dump(self):
        pass

    # verify entire dataset is correct in membase at our current position
    def verify(self):
        pass

    # get the current state (num ops, percent complete, time elapsed)
    # also get the number of failed ops and failed add/set (failed adds due to item existing don't count)
    # return total ops, op breakdown (gets, sets, etc), total ops/s and ops/s breakdown (gets/s, sets/s, etc)
    def query(self):
        for t in self.threads:
            t.join(0)
        self.threads = [t for t in self.threads if t.isAlive()]

        if not len(self.threads):
            self.stopped = True

        if self.stopped:
            state = 'stopped'
        elif self.paused:
            state = 'paused'
        else:
            state = 'running'

        return {
            'ops':0,
            'time':0,
            'keys':0,
            'state':state,
        }

    # block till condition
    def wait(self, time_limit=None):
        if time_limit == None:
            for t in self.threads:
                t.join()
        else:
            start_time = time.time()
            while len(self.threads) > 0:
                for t in self.threads:
                    if time.time() - start_time > time_limit:
                        return
                    t.join(1)
                self.threads = [t for t in self.threads if t.isAlive()]



from random import shuffle
import logger
from membase.api.rest_client import RestConnection, RestHelper
from remote.remote_util import RemoteMachineShellConnection

num_nodes_mismatch = "self.server has {0} nodes but cluster has {1} nodes"
not_enough_nodes = "there are not enough nodes to failover.cluster has{0} nodes and test wants to failover {1} nodes"
start_cluster = "rpc:call('{0}', ns_server_cluster_sup, start_cluster, [], infinity)."
stop_cluster = "rpc:call('{0}', ns_server_cluster_sup, stop_cluster, [], infinity)."

class FailoverHelper(object):
    def __init__(self, servers, test):
        self.log = logger.Logger.get_logger()
        self.servers = servers
        self.test = test
        #master is usually the first node ?

    # failover any node except self.servers[0]
    # assuming that replica = howmany
    def failover(self, howmany):
        #chekck if all nodes are part of the cluster
        rest = RestConnection(self.servers[0])
        nodes = rest.node_statuses()
        if len(nodes) != len(self.servers):
            self.test.fail(num_nodes_mismatch.format(len(self.servers), len(nodes)))
        if len(nodes) - howmany < 2:
            self.test.fail(num_nodes_mismatch.format(len(nodes), howmany))
        master_node = rest.get_nodes_self()
        #when selecting make sure we dont pick the master node
        selection=[n for n in nodes if n.id!=master_node.id]

        shuffle(selection)
        failed = selection[0:howmany]
        for f in failed:
            self.log.info("will fail over node : {0}".format(f.id))

        if len(nodes) / (1 + howmany) >= 1:
            self.test.assertTrue(RestHelper(rest).wait_for_replication(900),
                            msg="replication did not finish after 15 minutes")
            for f in failed:
                self._stop_server(f)
                self.log.info("10 seconds delay to wait for membase-server to shutdown")
            #wait for 5 minutes until node is down

            for f in failed:
                if f.port == 8091:
                    self.test.assertTrue(RestHelper(rest).wait_for_node_status(f, "unhealthy", 300),
                                msg="node status is not unhealthy even after waiting for 5 minutes")
                self.test.assertTrue(rest.fail_over(f.id), msg="failover did not complete")
                self.log.info("failed over node : {0}".format(f.id))
        return failed


    # Start and add the failovered nodes back to the cluster and rebalance it
    def undo_failover(self, failover_nodes):
        self.log.info("Add nodes back to the cluster: {0}".format(failover_nodes))
        rest = RestConnection(self.servers[0])

        self._start_servers(failover_nodes)
        rest.rebalance(otpNodes=set([node.id for node in rest.node_statuses()] +
                                    [node.id for node in failover_nodes]),
                       ejectedNodes=[])
        rest.monitorRebalance()


    def _stop_server(self, node):
        master_rest = RestConnection(self.servers[0])
        for server in self.servers:
            rest = RestConnection(server)
            self.log.info("see if server {0}:{1} is running".format(server.ip, server.port))
            if not RestHelper(rest).is_ns_server_running(timeout_in_seconds=5):
                continue
            node_id = rest.get_nodes_self().id
            if node_id == node.id:
                # if its 8091 then do ssh otherwise use ns_servr
                if node.port == 8091:
                    shell = RemoteMachineShellConnection(server)
                    if shell.is_membase_installed():
                        shell.stop_membase()
                        self.log.info("Membase stopped")
                    else:
                        shell.stop_couchbase()
                        self.log.info("Couchbase stopped")
                    shell.disconnect()
                    break
                else:
                    self.log.info("running {0}".format(stop_cluster.format(node.id)))
                    master_rest.diag_eval(stop_cluster.format(node.id))


    def _start_servers(self, nodes):
        for node in nodes:
            for server in self.servers:
                if node.ip == server.ip and str(node.port) == server.port:
                    shell = RemoteMachineShellConnection(server)
                    shell.start_couchbase()


# Copyright 2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Timezone related utilities for BSON."""

from datetime import (timedelta,
                      tzinfo)

ZERO = timedelta(0)


class FixedOffset(tzinfo):
    """Fixed offset timezone, in minutes east from UTC.

    Implementation based from the Python `standard library documentation
    <http://docs.python.org/library/datetime.html#tzinfo-objects>`_.
    Defining __getinitargs__ enables pickling / copying.
    """

    def __init__(self, offset, name):
        if isinstance(offset, timedelta):
            self.__offset = offset
        else:
            self.__offset = timedelta(minutes=offset)
        self.__name = name

    def __getinitargs__(self):
        return self.__offset, self.__name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO


utc = FixedOffset(0, "UTC")
"""Fixed offset timezone representing UTC."""





class MembaseHttpExceptionTypes(object):

    UNAUTHORIZED = 1000
    NOT_REACHABLE = 1001
    NODE_ALREADY_JOINED = 1002
    NODE_CANT_ADD_TO_ITSELF = 1003
    BUCKET_CREATION_ERROR = 1004
    STATS_UNAVAILABLE = 1005
    REMOTE_CLUSTER_JOIN_FAILED = 1006


#base exception class for membase apis
class MembaseHttpException(Exception):
    def __init__(self):
        self._message = ""
        self.type = ""
        #you can embed the params values here
        #dictionary mostly
        self.parameters = dict()

    def __init__(self, message, type, parameters):
        self._message = message
        self.type = type
        #you can embed the params values here
        #dictionary mostly
        self.parameters = parameters

    def __str__(self):
        string = ''
        if self._message:
            string += self._message
        return string


class UnauthorizedException(MembaseHttpException):
    def __init__(self, username='', password=''):
        self._message = 'user not logged in'
        self.parameters = dict()
        self.parameters['username'] = username
        self.parameters['password'] = password
        self.type = MembaseHttpExceptionTypes.UNAUTHORIZED


class BucketCreationException(MembaseHttpException):
    def __init__(self, ip='', bucket_name=''):
        self.parameters = dict()
        self.parameters['host'] = ip
        self.parameters['bucket'] = bucket_name
        self.type = MembaseHttpExceptionTypes.BUCKET_CREATION_ERROR
        self._message = 'unable to create bucket {0} on the host @ {1}'.\
            format(bucket_name, ip)


class StatsUnavailableException(MembaseHttpException):
    def __init__(self):
        self.type = MembaseHttpExceptionTypes.STATS_UNAVAILABLE
        self._message = 'unable to get stats'


class ServerUnavailableException(MembaseHttpException):
    def __init__(self, ip=''):
        self.parameters = dict()
        self.parameters['host'] = ip
        self.type = MembaseHttpExceptionTypes.NOT_REACHABLE
        self._message = 'unable to reach the host @ {0}'.format(ip)


class InvalidArgumentException(MembaseHttpException):
    def __init__(self, api, parameters):
        self.parameters = parameters
        self.api = api
        self._message = '{0} failed when invoked with parameters: {1}'.\
            format(self.api, self.parameters)


class ServerSelfJoinException(MembaseHttpException):
    def __init__(self, nodeIp='', remoteIp=''):
        self._message = 'node: {0} already added to this cluster:{1}'.\
            format(remoteIp, nodeIp)
        self.parameters = dict()
        self.parameters['nodeIp'] = nodeIp
        self.parameters['remoteIp'] = remoteIp
        self.type = MembaseHttpExceptionTypes.NODE_CANT_ADD_TO_ITSELF


class ClusterRemoteException(MembaseHttpException):
    def __init__(self, nodeIp='', remoteIp=''):
        self._message = 'unable to add remote cluster : {0}:{1}'.\
            format(remoteIp, nodeIp)
        self.parameters = dict()
        self.parameters['nodeIp'] = nodeIp
        self.parameters['remoteIp'] = remoteIp
        self.type = MembaseHttpExceptionTypes.REMOTE_CLUSTER_JOIN_FAILED


class ServerAlreadyJoinedException(MembaseHttpException):
    def __init__(self, nodeIp='', remoteIp=''):
        self._message = 'node: {0} already added to this cluster:{1}'.\
            format(remoteIp, nodeIp)
        self.parameters = dict()
        self.parameters['nodeIp'] = nodeIp
        self.parameters['remoteIp'] = remoteIp
        self.type = MembaseHttpExceptionTypes.NODE_ALREADY_JOINED


class RebalanceFailedException(MembaseHttpException):
    def __init__(self, string=''):
        self._message = 'Rebalance Failed: {0}'.format(string)

class FailoverFailedException(MembaseHttpException):
    def __init__(self, string=''):
        self._message = 'Failover Node failed :{0} '.format(string)


class DesignDocCreationException(MembaseHttpException):
    def __init__(self, design_doc_name, reason=''):
        self._message = 'Error occured design document %s: %s' % (design_doc_name, reason)

class QueryViewException(MembaseHttpException):
    def __init__(self, view_name, reason=''):
        self._message = 'Error occured querying view %s: %s' % (view_name, reason)

class ReadDocumentException(MembaseHttpException):
      def __init__(self, doc_id, reason=''):
        self._message = 'Error occured looking up document %s: %s' % (doc_id, reason)

class CompactViewFailed(MembaseHttpException):
      def __init__(self, design_doc_name, reason=''):
        self._message = 'Error occured triggering compaction for design_doc %s: %s' % \
            (design_doc_name, reason)

class SetViewInfoNotFound(MembaseHttpException):
      def __init__(self, design_doc_name, reason=''):
        self._message = 'Error occured reading set_view _info of ddoc %s: %s' % \
            (design_doc_name, reason)


class GetBucketInfoFailed(MembaseHttpException):
    def __init__(self, bucket, reason=''):
        self._message = 'Error occured getting bucket information %s: %s' % \
            (bucket, reason)

class AddNodeException(MembaseHttpException):
    def __init__(self, nodeIp='', remoteIp='', reason=''):
        self._message = 'Error adding node: {0} to the cluster:{1} - {2}'.\
            format(remoteIp, nodeIp, reason)
        self.parameters = dict()
        self.parameters['nodeIp'] = nodeIp
        self.parameters['remoteIp'] = remoteIp



#!/usr/bin/env python
import logging
import json
import time
import multiprocessing
import Queue

from lib.membase.api.rest_client import RestConnection
from lib.membase.api.exception import ServerUnavailableException

from tabula.table import Table
from tabula.section import Section
from seriesly import Seriesly
import seriesly.exceptions

from metadata.visit import retrieve_meta

from paint import TABULA_CONV_FUNCS, TABULA_DECO_FUNCS
from server import Server
from mc_source import MemcachedSource
from mc_collector import MemcachedCollector
from json_handler import JsonHandler
from seriesly_handler import SerieslyHandler
from data_helper import DataHelper

# configuration for each tabula.section
SECTION_CONFIG = {"settings": {"id": 0,
                               "show_row_hdrs": False,
                               "show_col_hdrs": False,
                               "show_col_hdr_in_cell": True},
                  "storage": {"id": 1,
                              "show_row_hdrs": False,
                              "show_col_hdrs": False,
                              "show_col_hdr_in_cell": True},
                  "buckets": {"id": 2,
                            "show_row_hdrs": False,
                            "show_col_hdrs": True,
                            "show_col_hdr_in_cell": False},
                  "nodes": {"id": 3,
                            "show_row_hdrs": False,
                            "show_col_hdrs": True,
                            "show_col_hdr_in_cell": False},
                  "Memory Stats": {"id": 4,
                                     "show_row_hdrs": False,
                                     "show_col_hdrs": True,
                                     "show_col_hdr_in_cell": False}}

tbl = Table("cbtop", sep=" ")
cur_row = {}      # {sec_nam: row name}
mc_jobs = multiprocessing.Queue(1)
mc_stats = multiprocessing.Queue(20)
store = None

class SerieslyStore(object):

    def __init__(self, host, dbslow, dbfast):
        self.slow = {}
        self.fast = {}
        self.dbslow = dbslow
        self.dbfast = dbfast
        self.seriesly = Seriesly(host=host)
        try:
            dbs = self.seriesly.list_dbs()
        except seriesly.exceptions.ConnectionError, e:
            logging.error("unable to connect to seriesly server: %s" % e)
            return

        if dbslow not in dbs:
            self.seriesly.create_db(dbslow)
        if dbfast not in dbs:
            self.seriesly.create_db(dbfast)

    def clear(self):
        self.slow = {}
        self.fast = {}

    def add_fast(self, key, val):
        self.fast[key] = val

    def add_slow(self, key, val):
        self.slow[key] = val

    def persist(self):
        try:
            if self.slow:
                self.seriesly[self.dbslow].append(self.slow)
            if self.fast:
                self.seriesly[self.dbfast].append(self.fast)
        except seriesly.exceptions.ConnectionError, e:
            logging.error("unable to connect to seriesly server: %s" % e)
            return False

        return True

def _show_stats(key, val, meta_inf):
    """
    Show stats on the ascii table
    """
    if not tbl or not isinstance(tbl, Table):
        return False

    if not meta_inf or not "section" in meta_inf:
        logging.debug("unable to show data: key %s, val %s, invalid meta info"
                      % (key, val))
        return False

    # ok, not deal with unicode for now
    sec_nam = str(meta_inf["section"])
    val = str(val)

    section = tbl.get_section(sec_nam)

    if not section:
        if sec_nam in SECTION_CONFIG:
            config = SECTION_CONFIG[sec_nam]
            section = Section(sec_nam, config["id"],
                              conv_funcs=TABULA_CONV_FUNCS,
                              deco_funcs=TABULA_DECO_FUNCS)
            section.config(config["show_row_hdrs"],
                config["show_col_hdrs"],
                config["show_col_hdr_in_cell"])
        else:
            return False
        tbl.add_section(section)

    if "col" in meta_inf:
        col = str(meta_inf["col"])
    else:
        col = str(key)

    if "new_row" in meta_inf:
        # create a new row using the col name
        section.add_cell(val, col, val, "S50", meta=meta_inf)
        cur_row[sec_nam] = val
        return True

    if not sec_nam in cur_row:
        logging.debug("stats %s is not associated with a section" % key)
        return True

    row = cur_row[sec_nam]
    section.add_cell(row, col, val, type="S50", meta=meta_inf)

    return True

def show_all_stats(stats, meta):
    if not isinstance(stats, dict) or not isinstance(meta, dict):
        logging.error("failed to show all stats : invalid data")
        return False

    for key, val in stats.iteritems():
        if not key in meta:
            continue

        _show_stats(key, val, meta[key])

def store_fast(root, parents, data, meta, coll,
               key, val, meta_val, meta_inf, level):
    """Store time-series data into fast-changing database"""
    store.add_fast(key, val)
    return _show_stats(key, val, meta_inf)

def store_slow(root, parents, data, meta, coll,
               key, val, meta_val, meta_inf, level):
    """Store time-series data into slow-changing database"""
    store.add_slow(key, val)
    return _show_stats(key, val, meta_inf)

def url_before(context, path):
    return context, path

def url_after(context, path, root):
    pass

def retrieve_data(context, path):
    """Retrieve json data from a couchbase server through REST calls"""
    # TODO: use cbtestlib
    server = Server(context.get("host", "127.0.0.1"))
    rest = RestConnection(server)
    api = rest.baseUrl + path

    try:
        status, content, header = rest._http_request(api)   #TODO: expose
    except ServerUnavailableException, e:
        logging.error("unable to retrieve data from %s: %s" % (server, e))
        return retrieve_meta(context, path)

    if status:
        return json.loads(content)

    return retrieve_meta(context, path)

def collect_mc_stats(root, parents, data, meta, coll,
                     key, val, meta_val, meta_inf, level=0):
    """
    Collect memcached stats
    Dump time series data a json snapshot
    """
    if not isinstance(val, list):
        logging.error(
            "unable to collect mc stats: val must be a list - %s" % val)
        return False

    try:
        stats, meta = mc_stats.get(block=False)
        show_all_stats(stats, meta)
    except Queue.Empty:
        pass

    try:
        mc_jobs.put([root, parents, val], block=False)
        return True
    except Queue.Full:
        logging.debug("unable to collect mcstats : queue is full")
        return False

def mc_worker(jobs, stats, ctl, store, timeout=5):
    logging.info("mc_worker started")

    while ctl["run_ok"]:
        try:
            root, parents, val = jobs.get(block=True, timeout=timeout)
        except Queue.Empty:
            logging.debug("mc_worker hasn't received jobs for %s seconds"
                          % timeout)
            continue

        start = time.time()

        for server in val:

            try:
                ip, port = server.split(":")
            except (ValueError, AttributeError), e:
                logging.error("unable to collect mc stats from %s : %s"
                              % (server, e))
                continue

            mc_server = Server(ip)

            # get bucket name from root and parent nodes
            bucket = DataHelper.get_bucket(root, parents)

            # initialize memcached source
            mc_source = MemcachedSource(mc_server, bucket)

            # initialize handlers to dump data json doc
            j_handler = JsonHandler()
            s_hanlder = SerieslyHandler(store)

            # collect data from source and emit to handlers
            mc_coll = MemcachedCollector([mc_source], [j_handler, s_hanlder])
            mc_coll.collect()
            mc_coll.emit()
            stats.put([mc_source.fast, mc_source.meta], block=True)
            stats.put([mc_source.slow, mc_source.meta], block=True)

        delta = time.time() - start
        logging.debug("collected mc stats from %s, took %s seconds"
                      % (val, delta))

        if delta < timeout:
            logging.debug("mc_worker sleep for %s seconds" % (timeout -delta))
            time.sleep(timeout - delta)

    logging.info("mc_worker stopped")



# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Database level operations."""

import warnings

from bson.binary import OLD_UUID_SUBTYPE
from bson.code import Code
from bson.dbref import DBRef
from bson.son import SON
from pymongo import common, helpers
from pymongo.collection import Collection
from pymongo.errors import (CollectionInvalid,
                            InvalidName,
                            OperationFailure)
from pymongo.son_manipulator import ObjectIdInjector


def _check_name(name):
    """Check if a database name is valid.
    """
    if not name:
        raise InvalidName("database name cannot be the empty string")

    for invalid_char in [" ", ".", "$", "/", "\\"]:
        if invalid_char in name:
            raise InvalidName("database names cannot contain the "
                              "character %r" % invalid_char)


class Database(common.BaseObject):
    """A Mongo database.
    """

    def __init__(self, connection, name):
        """Get a database by connection and name.

        Raises :class:`TypeError` if `name` is not an instance of
        :class:`basestring`. Raises
        :class:`~pymongo.errors.InvalidName` if `name` is not a valid
        database name.

        :Parameters:
          - `connection`: a :class:`~pymongo.connection.Connection`
            instance
          - `name`: database name

        .. mongodoc:: databases
        """
        super(Database,
              self).__init__(slave_okay=connection.slave_okay,
                             read_preference=connection.read_preference,
                             safe=connection.safe,
                             **(connection.get_lasterror_options()))

        if not isinstance(name, basestring):
            raise TypeError("name must be an instance of basestring")

        _check_name(name)

        self.__name = unicode(name)
        self.__connection = connection

        self.__incoming_manipulators = []
        self.__incoming_copying_manipulators = []
        self.__outgoing_manipulators = []
        self.__outgoing_copying_manipulators = []
        self.add_son_manipulator(ObjectIdInjector())

    def add_son_manipulator(self, manipulator):
        """Add a new son manipulator to this database.

        Newly added manipulators will be applied before existing ones.

        :Parameters:
          - `manipulator`: the manipulator to add
        """
        def method_overwritten(instance, method):
            return getattr(instance, method) != \
                getattr(super(instance.__class__, instance), method)

        if manipulator.will_copy():
            if method_overwritten(manipulator, "transform_incoming"):
                self.__incoming_copying_manipulators.insert(0, manipulator)
            if method_overwritten(manipulator, "transform_outgoing"):
                self.__outgoing_copying_manipulators.insert(0, manipulator)
        else:
            if method_overwritten(manipulator, "transform_incoming"):
                self.__incoming_manipulators.insert(0, manipulator)
            if method_overwritten(manipulator, "transform_outgoing"):
                self.__outgoing_manipulators.insert(0, manipulator)

    @property
    def system_js(self):
        """A :class:`SystemJS` helper for this :class:`Database`.

        See the documentation for :class:`SystemJS` for more details.

        .. versionadded:: 1.5
        """
        return SystemJS(self)

    @property
    def connection(self):
        """The :class:`~pymongo.connection.Connection` instance for this
        :class:`Database`.

        .. versionchanged:: 1.3
           ``connection`` is now a property rather than a method.
        """
        return self.__connection

    @property
    def name(self):
        """The name of this :class:`Database`.

        .. versionchanged:: 1.3
           ``name`` is now a property rather than a method.
        """
        return self.__name

    @property
    def incoming_manipulators(self):
        """List all incoming SON manipulators
        installed on this instance.

        .. versionadded:: 2.0
        """
        return [manipulator.__class__.__name__
                for manipulator in self.__incoming_manipulators]

    @property
    def incoming_copying_manipulators(self):
        """List all incoming SON copying manipulators
        installed on this instance.

        .. versionadded:: 2.0
        """
        return [manipulator.__class__.__name__
                for manipulator in self.__incoming_copying_manipulators]

    @property
    def outgoing_manipulators(self):
        """List all outgoing SON manipulators
        installed on this instance.

        .. versionadded:: 2.0
        """
        return [manipulator.__class__.__name__
                for manipulator in self.__outgoing_manipulators]

    @property
    def outgoing_copying_manipulators(self):
        """List all outgoing SON copying manipulators
        installed on this instance.

        .. versionadded:: 2.0
        """
        return [manipulator.__class__.__name__
                for manipulator in self.__outgoing_copying_manipulators]

    def __cmp__(self, other):
        if isinstance(other, Database):
            return cmp((self.__connection, self.__name),
                       (other.__connection, other.__name))
        return NotImplemented

    def __repr__(self):
        return "Database(%r, %r)" % (self.__connection, self.__name)

    def __getattr__(self, name):
        """Get a collection of this database by name.

        Raises InvalidName if an invalid collection name is used.

        :Parameters:
          - `name`: the name of the collection to get
        """
        return Collection(self, name)

    def __getitem__(self, name):
        """Get a collection of this database by name.

        Raises InvalidName if an invalid collection name is used.

        :Parameters:
          - `name`: the name of the collection to get
        """
        return self.__getattr__(name)

    def create_collection(self, name, **kwargs):
        """Create a new :class:`~pymongo.collection.Collection` in this
        database.

        Normally collection creation is automatic. This method should
        only be used to specify options on
        creation. :class:`~pymongo.errors.CollectionInvalid` will be
        raised if the collection already exists.

        Options should be passed as keyword arguments to this
        method. Any of the following options are valid:

          - "size": desired initial size for the collection (in
            bytes). must be less than or equal to 10000000000. For
            capped collections this size is the max size of the
            collection.
          - "capped": if True, this is a capped collection
          - "max": maximum number of objects if capped (optional)

        :Parameters:
          - `name`: the name of the collection to create
          - `**kwargs` (optional): additional keyword arguments will
            be passed as options for the create collection command

        .. versionchanged:: 2.1.1+
           Removed deprecated argument: options

        .. versionchanged:: 1.5
           deprecating `options` in favor of kwargs
        """
        opts = {"create": True}
        opts.update(kwargs)

        if name in self.collection_names():
            raise CollectionInvalid("collection %s already exists" % name)

        return Collection(self, name, **opts)

    def _fix_incoming(self, son, collection):
        """Apply manipulators to an incoming SON object before it gets stored.

        :Parameters:
          - `son`: the son object going into the database
          - `collection`: the collection the son object is being saved in
        """
        for manipulator in self.__incoming_manipulators:
            son = manipulator.transform_incoming(son, collection)
        for manipulator in self.__incoming_copying_manipulators:
            son = manipulator.transform_incoming(son, collection)
        return son

    def _fix_outgoing(self, son, collection):
        """Apply manipulators to a SON object as it comes out of the database.

        :Parameters:
          - `son`: the son object coming out of the database
          - `collection`: the collection the son object was saved in
        """
        for manipulator in reversed(self.__outgoing_manipulators):
            son = manipulator.transform_outgoing(son, collection)
        for manipulator in reversed(self.__outgoing_copying_manipulators):
            son = manipulator.transform_outgoing(son, collection)
        return son

    def command(self, command, value=1,
                check=True, allowable_errors=[],
                uuid_subtype=OLD_UUID_SUBTYPE, **kwargs):
        """Issue a MongoDB command.

        Send command `command` to the database and return the
        response. If `command` is an instance of :class:`basestring`
        then the command {`command`: `value`} will be sent. Otherwise,
        `command` must be an instance of :class:`dict` and will be
        sent as is.

        Any additional keyword arguments will be added to the final
        command document before it is sent.

        For example, a command like ``{buildinfo: 1}`` can be sent
        using:

        >>> db.command("buildinfo")

        For a command where the value matters, like ``{collstats:
        collection_name}`` we can do:

        >>> db.command("collstats", collection_name)

        For commands that take additional arguments we can use
        kwargs. So ``{filemd5: object_id, root: file_root}`` becomes:

        >>> db.command("filemd5", object_id, root=file_root)

        :Parameters:
          - `command`: document representing the command to be issued,
            or the name of the command (for simple commands only).

            .. note:: the order of keys in the `command` document is
               significant (the "verb" must come first), so commands
               which require multiple keys (e.g. `findandmodify`)
               should use an instance of :class:`~bson.son.SON` or
               a string and kwargs instead of a Python `dict`.

          - `value` (optional): value to use for the command verb when
            `command` is passed as a string
          - `check` (optional): check the response for errors, raising
            :class:`~pymongo.errors.OperationFailure` if there are any
          - `allowable_errors`: if `check` is ``True``, error messages
            in this list will be ignored by error-checking
          - `uuid_subtype` (optional): The BSON binary subtype to use
            for a UUID used in this command.
          - `**kwargs` (optional): additional keyword arguments will
            be added to the command document before it is sent

        .. versionchanged:: 1.6
           Added the `value` argument for string commands, and keyword
           arguments for additional command options.
        .. versionchanged:: 1.5
           `command` can be a string in addition to a full document.
        .. versionadded:: 1.4

        .. mongodoc:: commands
        """

        if isinstance(command, basestring):
            command = SON([(command, value)])

        extra_opts = {
            'read_preference': kwargs.pop('read_preference',
                                          self.read_preference),
            'slave_okay': kwargs.pop('slave_okay', self.slave_okay),
            '_must_use_master': kwargs.pop('_use_master', True),
            '_is_command': True,
            '_uuid_subtype': uuid_subtype
        }

        fields = kwargs.get('fields')
        if fields is not None and not isinstance(fields, dict):
                kwargs['fields'] = helpers._fields_list_to_dict(fields)

        command.update(kwargs)

        result = self["$cmd"].find_one(command, **extra_opts)

        if check:
            msg = "command %s failed: %%s" % repr(command).replace("%", "%%")
            helpers._check_command_response(result, self.connection.disconnect,
                                            msg, allowable_errors)

        return result

    def collection_names(self):
        """Get a list of all the collection names in this database.
        """
        results = self["system.namespaces"].find(_must_use_master=True)
        names = [r["name"] for r in results]
        names = [n[len(self.__name) + 1:] for n in names
                 if n.startswith(self.__name + ".")]
        names = [n for n in names if "$" not in n]
        return names

    def drop_collection(self, name_or_collection):
        """Drop a collection.

        :Parameters:
          - `name_or_collection`: the name of a collection to drop or the
            collection object itself
        """
        name = name_or_collection
        if isinstance(name, Collection):
            name = name.name

        if not isinstance(name, basestring):
            raise TypeError("name_or_collection must be an instance of "
                            "(Collection, str, unicode)")

        self.__connection._purge_index(self.__name, name)

        self.command("drop", unicode(name), allowable_errors=["ns not found"])

    def validate_collection(self, name_or_collection,
                            scandata=False, full=False):
        """Validate a collection.

        Returns a dict of validation info. Raises CollectionInvalid if
        validation fails.

        With MongoDB < 1.9 the result dict will include a `result` key
        with a string value that represents the validation results. With
        MongoDB >= 1.9 the `result` key no longer exists and the results
        are split into individual fields in the result dict.

        :Parameters:
            `name_or_collection`: A Collection object or the name of a
                                  collection to validate.
            `scandata`: Do extra checks beyond checking the overall
                        structure of the collection.
            `full`: Have the server do a more thorough scan of the
                    collection. Use with `scandata` for a thorough scan
                    of the structure of the collection and the individual
                    documents. Ignored in MongoDB versions before 1.9.

        .. versionchanged:: 1.11
           validate_collection previously returned a string.
        .. versionadded:: 1.11
           Added `scandata` and `full` options.
        """
        name = name_or_collection
        if isinstance(name, Collection):
            name = name.name

        if not isinstance(name, basestring):
            raise TypeError("name_or_collection must be an instance of "
                            "(Collection, str, unicode)")

        result = self.command("validate", unicode(name),
                              scandata=scandata, full=full)

        valid = True
        # Pre 1.9 results
        if "result" in result:
            info = result["result"]
            if info.find("exception") != -1 or info.find("corrupt") != -1:
                raise CollectionInvalid("%s invalid: %s" % (name, info))
        # Sharded results
        elif "raw" in result:
            for repl, res in result["raw"].iteritems():
                if "result" in res:
                    info = res["result"]
                    if (info.find("exception") != -1 or
                        info.find("corrupt") != -1):
                        raise CollectionInvalid("%s invalid: "
                                                "%s" % (name, info))
                elif not res.get("valid", False):
                    valid = False
                    break
        # Post 1.9 non-sharded results.
        elif not result.get("valid", False):
            valid = False

        if not valid:
            raise CollectionInvalid("%s invalid: %r" % (name, result))

        return result

    def current_op(self):
        """Get information on operations currently running.
        """
        return self['$cmd.sys.inprog'].find_one()

    def profiling_level(self):
        """Get the database's current profiling level.

        Returns one of (:data:`~pymongo.OFF`,
        :data:`~pymongo.SLOW_ONLY`, :data:`~pymongo.ALL`).

        .. mongodoc:: profiling
        """
        result = self.command("profile", -1)

        assert result["was"] >= 0 and result["was"] <= 2
        return result["was"]

    def set_profiling_level(self, level):
        """Set the database's profiling level.

        Raises :class:`ValueError` if level is not one of
        (:data:`~pymongo.OFF`, :data:`~pymongo.SLOW_ONLY`,
        :data:`~pymongo.ALL`).

        :Parameters:
          - `level`: the profiling level to use

        .. mongodoc:: profiling
        """
        if not isinstance(level, int) or level < 0 or level > 2:
            raise ValueError("level must be one of (OFF, SLOW_ONLY, ALL)")

        self.command("profile", level)

    def profiling_info(self):
        """Returns a list containing current profiling information.

        .. mongodoc:: profiling
        """
        return list(self["system.profile"].find())

    def error(self):
        """Get a database error if one occured on the last operation.

        Return None if the last operation was error-free. Otherwise return the
        error that occurred.
        """
        error = self.command("getlasterror")
        error_msg = error.get("err", "")
        if error_msg is None:
            return None
        if error_msg.startswith("not master"):
            self.__connection.disconnect()
        return error

    def last_status(self):
        """Get status information from the last operation.

        Returns a SON object with status information.
        """
        return self.command("getlasterror")

    def previous_error(self):
        """Get the most recent error to have occurred on this database.

        Only returns errors that have occurred since the last call to
        `Database.reset_error_history`. Returns None if no such errors have
        occurred.
        """
        error = self.command("getpreverror")
        if error.get("err", 0) is None:
            return None
        return error

    def reset_error_history(self):
        """Reset the error history of this database.

        Calls to `Database.previous_error` will only return errors that have
        occurred since the most recent call to this method.
        """
        self.command("reseterror")

    def __iter__(self):
        return self

    def next(self):
        raise TypeError("'Database' object is not iterable")

    def add_user(self, name, password, read_only=False):
        """Create user `name` with password `password`.

        Add a new user with permissions for this :class:`Database`.

        .. note:: Will change the password if user `name` already exists.

        :Parameters:
          - `name`: the name of the user to create
          - `password`: the password of the user to create
          - `read_only` (optional): if ``True`` it will make user read only

        .. versionchanged:: 2.1.1+
           Added support for read only users

        .. versionadded:: 1.4
        """
        pwd = helpers._password_digest(name, password)
        self.system.users.update({"user": name},
                                 {"user": name,
                                  "pwd": pwd,
                                  "readOnly": read_only},
                                 upsert=True, safe=True)

    def remove_user(self, name):
        """Remove user `name` from this :class:`Database`.

        User `name` will no longer have permissions to access this
        :class:`Database`.

        :Parameters:
          - `name`: the name of the user to remove

        .. versionadded:: 1.4
        """
        self.system.users.remove({"user": name}, safe=True)

    def authenticate(self, name, password):
        """Authenticate to use this database.

        Once authenticated, the user has full read and write access to
        this database. Raises :class:`TypeError` if either `name` or
        `password` is not an instance of ``(str,
        unicode)``. Authentication lasts for the life of the underlying
        :class:`~pymongo.connection.Connection`, or until :meth:`logout`
        is called.

        The "admin" database is special. Authenticating on "admin"
        gives access to *all* databases. Effectively, "admin" access
        means root access to the database.

        .. note::
          This method authenticates the current connection, and
          will also cause all new :class:`~socket.socket` connections
          in the underlying :class:`~pymongo.connection.Connection` to
          be authenticated automatically.

         - When sharing a :class:`~pymongo.connection.Connection`
           between multiple threads, all threads will share the
           authentication. If you need different authentication profiles
           for different purposes (e.g. admin users) you must use
           distinct instances of :class:`~pymongo.connection.Connection`.

         - To get authentication to apply immediately to all
           existing sockets you may need to reset this Connection's
           sockets using :meth:`~pymongo.connection.Connection.disconnect`.

        .. warning::

          Currently, calls to
          :meth:`~pymongo.connection.Connection.end_request` will
          lead to unpredictable behavior in combination with
          auth. The :class:`~socket.socket` owned by the calling
          thread will be returned to the pool, so whichever thread
          uses that :class:`~socket.socket` next will have whatever
          permissions were granted to the calling thread.

        :Parameters:
          - `name`: the name of the user to authenticate
          - `password`: the password of the user to authenticate

        .. mongodoc:: authenticate
        """
        if not isinstance(name, basestring):
            raise TypeError("name must be an instance of basestring")
        if not isinstance(password, basestring):
            raise TypeError("password must be an instance of basestring")

        nonce = self.command("getnonce")["nonce"]
        key = helpers._auth_key(nonce, name, password)
        try:
            self.command("authenticate", user=unicode(name),
                         nonce=nonce, key=key)
            self.connection._cache_credentials(self.name,
                                               unicode(name),
                                               unicode(password))
            return True
        except OperationFailure:
            return False

    def logout(self):
        """Deauthorize use of this database for this connection
        and future connections.

        .. note:: Other databases may still be authenticated, and other
           existing :class:`~socket.socket` connections may remain
           authenticated for this database unless you reset all sockets
           with :meth:`~pymongo.connection.Connection.disconnect`.
        """
        self.command("logout")
        self.connection._purge_credentials(self.name)

    def dereference(self, dbref):
        """Dereference a :class:`~bson.dbref.DBRef`, getting the
        document it points to.

        Raises :class:`TypeError` if `dbref` is not an instance of
        :class:`~bson.dbref.DBRef`. Returns a document, or ``None`` if
        the reference does not point to a valid document.  Raises
        :class:`ValueError` if `dbref` has a database specified that
        is different from the current database.

        :Parameters:
          - `dbref`: the reference
        """
        if not isinstance(dbref, DBRef):
            raise TypeError("cannot dereference a %s" % type(dbref))
        if dbref.database is not None and dbref.database != self.__name:
            raise ValueError("trying to dereference a DBRef that points to "
                             "another database (%r not %r)" % (dbref.database,
                                                               self.__name))
        return self[dbref.collection].find_one({"_id": dbref.id})

    def eval(self, code, *args):
        """Evaluate a JavaScript expression in MongoDB.

        Useful if you need to touch a lot of data lightly; in such a
        scenario the network transfer of the data could be a
        bottleneck. The `code` argument must be a JavaScript
        function. Additional positional arguments will be passed to
        that function when it is run on the server.

        Raises :class:`TypeError` if `code` is not an instance of
        (str, unicode, `Code`). Raises
        :class:`~pymongo.errors.OperationFailure` if the eval
        fails. Returns the result of the evaluation.

        :Parameters:
          - `code`: string representation of JavaScript code to be
            evaluated
          - `args` (optional): additional positional arguments are
            passed to the `code` being evaluated
        """
        if not isinstance(code, Code):
            code = Code(code)

        result = self.command("$eval", code, args=args)
        return result.get("retval", None)

    def __call__(self, *args, **kwargs):
        """This is only here so that some API misusages are easier to debug.
        """
        raise TypeError("'Database' object is not callable. If you meant to "
                        "call the '%s' method on a 'Connection' object it is "
                        "failing because no such method exists." % self.__name)


class SystemJS(object):
    """Helper class for dealing with stored JavaScript.
    """

    def __init__(self, database):
        """Get a system js helper for the database `database`.

        An instance of :class:`SystemJS` can be created with an instance
        of :class:`Database` through :attr:`Database.system_js`,
        manual instantiation of this class should not be necessary.

        :class:`SystemJS` instances allow for easy manipulation and
        access to server-side JavaScript:

        .. doctest::

          >>> db.system_js.add1 = "function (x) { return x + 1; }"
          >>> db.system.js.find({"_id": "add1"}).count()
          1
          >>> db.system_js.add1(5)
          6.0
          >>> del db.system_js.add1
          >>> db.system.js.find({"_id": "add1"}).count()
          0

        .. note:: Requires server version **>= 1.1.1**

        .. versionadded:: 1.5
        """
        # can't just assign it since we've overridden __setattr__
        object.__setattr__(self, "_db", database)

    def __setattr__(self, name, code):
        self._db.system.js.save({"_id": name, "value": Code(code)}, safe=True)

    def __setitem__(self, name, code):
        self.__setattr__(name, code)

    def __delattr__(self, name):
        self._db.system.js.remove({"_id": name}, safe=True)

    def __delitem__(self, name):
        self.__delattr__(name)

    def __getattr__(self, name):
        return lambda *args: self._db.eval(Code("function() { "
                                                "return this[name].apply("
                                                "this, arguments); }",
                                                scope={'name': name}), *args)

    def __getitem__(self, name):
        return self.__getattr__(name)

    def list(self):
        """Get a list of the names of the functions stored in this database.

        .. versionadded:: 1.9
        """
        return [x["_id"] for x in self._db.system.js.find(fields=["_id"])]



#!/usr/bin/env python

import logging
import os
import ast

import numpy as np

UNITS_SUFFIX = ["", "K", "M", "G", "T"]
TYPE_META = "S512"

def conv_units(val, meta):
    """
    Format and convert units to be more human readable

    @return new val with converted units
    """
    if not val or not meta:
        return val

    try:
        val = float(val)
    except ValueError:
        logging.error("unable to apply convert units for %s" % val)
        return val

    suf = 0
    while val > 1024 and suf < 4:
        val /= 1024
        suf += 1

    return "%.2f%s" % (val, UNITS_SUFFIX[suf])

"""CONV_FUNCS: convert data before column alignments"""
CONV_FUNCS = {"units": conv_units}

"""DECO_FUNCS: decorate data after column alignments"""
DECO_FUNCS = {}

class Section(object):
    """
    A section appears on @class Table

    e.g:

    - Flat section:
        key1: val1, key2: val2
        key3: val3 ...

    - Grid section:
        Like a spreedsheet, with col and row headers

        name |  col1 |  col2
        row1 |  val1 |  val2
        row2 |  val3 |  val4

    """
    def __init__(self, name, id=0, width=800, height=600, sep="  ",
                 show_row_hdrs=True, show_col_hdrs=True,
                 show_col_hdr_in_cell=False, auto_resize=True,
                 conv_funcs=CONV_FUNCS, deco_funcs=DECO_FUNCS):
        """
        @param id: id/position to place a section on the table
        @param width : max width, use terminal width auto_resize is on
        @param height: max height

        @param show_row_hdrs : show row headers
        @param show_col_hdrs : show column headers
        @param show_col_hdr_in_cell : embed column header in each cell
        @param auto_resize : auto resize according to the size of terminal
        @param conv_funcs : func to convert data before column alignments
        @param deco_funcs : func to decorate data after column alignments
        """
        self.name = name
        self.id = id
        self.width = width
        self.height = height
        self.sep = sep
        self.show_row_hdrs = show_row_hdrs
        self.show_col_hdrs = show_col_hdrs
        self.show_col_hdr_in_cell = show_col_hdr_in_cell
        self.auto_resize = auto_resize
        self.conv_funcs = conv_funcs
        self.deco_funcs = deco_funcs
        self.arr = None
        self.meta = None
        self.irt = {}           # inverted-row-table @dict {row_name: row_num}
        self._create()

    def __repr__(self):
        return "Section(name=%r, width=%r, height=%r, sep=%r, "\
               "show_row_hdrs=%r, show_col_hdrs=%r, "\
               "show_col_hdr_in_cell=%r, arr=%r, irt=%r)"\
               % (self.name, self.width, self.height, self.sep,
                  self.show_row_hdrs, self.show_col_hdrs,
                  self.show_col_hdr_in_cell, self.arr, self.irt)

    def __str__(self):
        return self._format()[0]

    def _term_size(self):
        """
        Retrieve the size of the terminal

        @return [width, height]
        """
        return list(reversed(os.popen('stty size', 'r').read().split()))

    def align(self, arr):
        """
        Align columns, including column headers
        """
        if arr is None:
            return arr

        c_hdrs = self._get_col_hdrs()

        if self.show_col_hdr_in_cell:
            for hdr in c_hdrs:
                arr[hdr] = map(lambda col: ":".join([hdr, str(col)]), arr[hdr])

        if self.show_col_hdrs:
            widths = [max(len(str(col))
                      for col in arr[hdr].tolist() + [hdr]) for hdr in c_hdrs]
        else:
            widths = [max(len(str(col))
                      for col in arr[hdr].tolist()) for hdr in c_hdrs]

        # align column headers
        c_hdrs = map(lambda (c_hdr, width): c_hdr.ljust(width),
                     zip(c_hdrs, widths))

        # align data
        for n_row in range(len(arr)):
            arr[n_row] = tuple(map(lambda (col, width): col.ljust(width),
                                   zip(arr[n_row], widths)))

        return arr, c_hdrs, widths

    def _format(self):

        if self.arr is None:
            return "", 0, 0

        if self.auto_resize:
            self.width = int(self._term_size()[0])

        arr = self.apply_meta(self.arr, self.conv_funcs)
        arr = self.sort(arr)

        if arr is None:
            return ""

        arr, c_hdrs, widths = self.align(arr)
        arr = self.apply_meta(arr, self.deco_funcs)

        data = arr.tolist()
        height = len(data)

        if self.show_row_hdrs:
            height += 1
            if self.show_col_hdrs:
                string = self.sep.join(c_hdrs) + "\n"
            else:
                string = ""
            string += "\n".join(self.sep.join(col for col in row)
                                for row in data)
        else:
            widths = widths[1:]
            if self.show_col_hdrs:
                string = self.sep.join(c_hdrs[1:]) + "\n"
            else:
                string = ""
            string += "\n".join(self.sep.join(col for col in row[1:])
                                for row in data)

        return self.wrap(string, self.width),\
               min(self.width, sum(widths) + (len(widths) - 1) * len(self.sep)),\
               height

    def format(self):
        """
        Format the section to a presentable string

        See @func config() for how to override configurations

        @return (formatted string, viewable width, viewable height)
        """
        return self._format()

    def wrap(self, string, width):
        """
        Wrap lines according to width
        Place '\n' whenever necessary
        """
        if not string or width <= 0:
            logging.error("invalid string: %s or width: %s" % (string, width))
            return False

        tmp = ""
        for line in string.splitlines():
            if len(line) <= width:
                tmp += line + "\n"
                continue

            cur = 0
            length = len(line)
            while cur + width < length:
                cur = line[:cur+width].rfind(self.sep) + len(self.sep) - 1
                line = line[:cur] + "\n" + line[cur+1:]

            tmp += line + "\n\n"

        return tmp

    def apply_meta(self, arr, funcs):
        """
        Apply metadata to help formatting the output:

        - conv_funcs: convert data before column alignments
        - deco_funcs: decorate data after column alignments
        """
        if arr is None:
            logging.error("unable to convert data: emtpy section")
            return None

        tmp = np.copy(arr)
        for col in arr.dtype.names:
            for row in self._get_row_hdrs():
                meta = self._get_meta(row, col)
                for mk, mv in sorted(meta.iteritems()):
                    if mk in funcs.iterkeys():
                        tmp[col][self.irt[row]] = \
                            funcs[mk](tmp[col][self.irt[row]], mv)
        return tmp

    def config(self, show_row_hdrs=True, show_col_hdrs=True,
                show_col_hdr_in_cell=False, auto_resize=True):
        """
        Override the in-class params:

        @param show_row_hdrs : show row headers
        @param show_col_hdrs : show column headers
        @param show_col_hdr_in_cell : embed column header in each cell
        @param auto_resize : auto resize according to the size of terminal
        """
        self.show_row_hdrs = show_row_hdrs
        self.show_col_hdrs = show_col_hdrs
        self.show_col_hdr_in_cell = show_col_hdr_in_cell

    def size(self):
        """
        Return the viewable size of the section as @tuple (width, height)
        """
        return self._format()[1:]

    def sort(self, arr, col=None):
        if arr is None:
            logging.error("unable to sort empty section")
            return None

        if not col or col not in self._get_col_hdrs():
            return np.sort(arr, order=self._get_col_hdrs()[1:])
        else:
            return np.sort(arr, order=col)

    def add_cell(self, row="unknown", col="unknown",
                 val="unknown", type="int32", meta=""):
        """
        Add/update a val on cell [row, col]

        Create new rows or columns accordingly

        @param row : row header name
        @param col : column header name
        @param meta: meta data to control data display
        """
        if self.arr is None:
            self.arr = np.array(
                [(row, val)], dtype=[ (self.name, "S50"), (col, type)])
            self.meta = np.array(
                [(row, val)], dtype=[ (self.name, TYPE_META), (col, type)])
            self.irt[row] = 0

        if not row in self._get_row_hdrs():
            self._expand_row(row)

        if not col in self._get_col_hdrs():
            self._expand_col(col, type)

        try:
            self._add_cell(val, row, col, meta)
        except ValueError:
            logging.error("unable to add val %s to [%s,%s]: "
                          "not a compatible data type" % (val, row, col))
            return False

        return True

    def get_cell(self, row, col):
        #TODO
        raise NotImplementedError

    def add_row(self, name, vals):
        #TODO
        raise NotImplementedError

    def del_row(self, name):

        if self.arr is None:
            logging.error("unable to delete row %s: empty section" % name)
            return False

        if not name in self._get_row_hdrs() or\
           not name in self.irt:
            logging.error("unable to delete row %s: row doesn't exist" % name)
            return False

        row_num = self.irt[name]
        self.arr = np.delete(self.arr, row_num)
        self.meta = np.delete(self.meta, row_num)
        self.irt.update(
            {k:v-1 for k,v in self.irt.iteritems() if v > row_num})

        return True

    def _create(self):

        if not self.arr is None:
            logging.error(
                "unable to create table %s: already exist" % self.name)
            return False

        self.arr = np.array([], dtype=[(self.name, "S50")])
        self.meta = np.array([], dtype=[(self.name, TYPE_META)])

        return True

    def _get_meta(self, row, col):
        """
        Get metadata for a particular cell
        """
        if self.meta is None:
            logging.error("unable to get meta: empty section")
            return {}

        if not row in self._get_row_hdrs() or\
            not col in self._get_col_hdrs():
            logging.error("unable to get meta: cell [%s,%s] does not exist"
                          % (row, col))
            return {}

        meta_str = self.meta[col][self.irt[row]]
        try:
            meta = ast.literal_eval(meta_str)
            if isinstance(meta, dict):
                return meta
        except (SyntaxError, ValueError), e:
            logging.error("unable to parse meta string - %s: %s"
                          % (meta_str, e))

        return {}

    def _get_col_hdrs(self):

        if self.arr is None:
            logging.error("unable to get row headers: empty section")
            return ()

        return self.arr.dtype.names

    def _get_row_hdrs(self):

        if self.arr is None:
            logging.error("unable to get row headers: empty section")
            return ()

        return self.arr[self.name]

    def _expand_col(self, name, type="int32"):

        if self.arr is None:
            logging.error("unable to add column %s: empty section" % name)
            return False

        if name in self._get_col_hdrs():
            logging.error("unable to add column %s: already exist" % name)
            return False

        arr_dtype = self.arr.dtype.descr + [(name, type)]
        new_arr = np.zeros(self.arr.shape, dtype=arr_dtype)
        meta_dtype = self.meta.dtype.descr + [(name, TYPE_META)]
        new_meta = np.zeros(self.meta.shape, dtype=meta_dtype)

        for field in self.arr.dtype.fields:
            new_arr[field] = self.arr[field]
        for field in self.meta.dtype.fields:
            new_meta[field] = self.meta[field]

        self.arr = new_arr
        self.meta = new_meta
        return True

    def _expand_row(self, name):

        if self.arr is None:
            logging.error("unable to add row %s: empty section" % name)
            return False

        if name in self._get_row_hdrs():
            logging.error("unable to add row %s: already exist" % name)
            return False

        n_rows = len(self.arr)
        self.arr = np.insert(self.arr, n_rows, np.array([name,]), 0)
        self.meta = np.insert(self.meta, n_rows, np.array([name,]), 0)

        self.irt[name] = n_rows

        return True

    def _add_cell(self, val, row, col, meta=""):
        """
        @except ValueError : data type not compatible
        """
        if self.arr is None:
            logging.error("unable to add value %s to [%s,%s]: empty section"
                          % (val, row, col))
            return False

        if not row in self._get_row_hdrs() or\
                not row in self.irt:
            logging.error("unable to add value %s to [%s,%s]: row doesn't exist"
                          % (val, row, col))
            return False

        if not col in self._get_col_hdrs():
            logging.error(
                "unable to add value %s to [%s,%s]: column doesn't exist"
                % (val, row, col))
            return False

        self.arr[self.irt[row]][col] = val
        self.meta[self.irt[row]][col] = str(meta)

        return True


#!/usr/bin/env python

from sys import path
from struct import calcsize

path.append("lib")

from memcacheConstants import REQ_PKT_FMT

OBS_OPCODE = 0x92

class ObserveKeyState:
    OBS_NOT_FOUND = 0x80
    OBS_FOUND = 0x00
    OBS_PERSISITED = 0x01
    OBS_IMPOSSIBLE = 0x81

class ObserveStatus:
    OBS_SUCCESS = 0x00
    OBS_MODIFIED = 0x01
    OBS_TIMEDOUT = 0x02
    OBS_UNKNOWN = 0xff
    OBS_ERROR = 0xee

class ObservePktFmt:
    OBS_REQ_HDR_FMT = REQ_PKT_FMT
    OBS_REQ_BODY_FMT = "!HH"
    OBS_RES_HDR_FMT = "!BBHBBHIIII"
    OBS_RES_BODY_FMT = "!HH"
    OBS_RES_HDR_LEN = calcsize(OBS_RES_HDR_FMT)
    OBS_RES_BODY_LEN = calcsize(OBS_RES_BODY_FMT)


#!/usr/bin/env python
"""
Ascii memcached test client.
"""

import socket
import select
import exceptions

import memcacheConstants

class MemcachedError(exceptions.Exception):
    """Error raised when a command fails."""

    def __init__(self, status, msg):
        supermsg='Memcached error #' + `status`
        if msg: supermsg += ":  " + msg
        exceptions.Exception.__init__(self, supermsg)

        self.status=status
        self.msg=msg

    def __repr__(self):
        return "<MemcachedError #%d ``%s''>" % (self.status, self.msg)

class MemcachedAsciiClient(object):
    """Simple ascii memcached client."""

    def __init__(self, host='127.0.0.1', port=11211, timeout=30):
        self.host = host
        self.port = port
        self.s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.timeout = timeout
        self.s.connect_ex((host, port))

    def close(self):
        self.s.close()

    def __del__(self):
        self.close()

    def _sendMsg(self, cmd):
        _, w, _ = select.select([], [self.s], [], self.timeout)
        if w:
            self.s.send(cmd)
        else:
            raise exceptions.EOFError("Timeout waiting for socket send. from {0}".format(self.host))

    def _recvMsg(self):
        r, _, _ = select.select([self.s], [], [], self.timeout)
        if r:
            response = ""
            while not response.endswith("\r\n"):
                data = self.s.recv(1)
                if data == '':
                    raise exceptions.EOFError("Got empty data (remote died?). from {0}".format(self.host))
                response += data
            return response[:-2]
        else:
            raise exceptions.EOFError("Timeout waiting for socket recv. from {0}".format(self.host))

    def _recvData(self, length):
        r, _, _ = select.select([self.s], [], [], self.timeout)
        if r:
            response = ""
            while len(response) < length + 2:
                data = self.s.recv((length + 2) - len(response))
                if data == '':
                    raise exceptions.EOFError("Got empty data (remote died?). from {0}".format(self.host))
                response += data
            return response[:-2]
        else:
            raise exceptions.EOFError("Timeout waiting for socket recv. from {0}".format(self.host))

    def _doStore(self, cmd):
        """Send a command and await its response."""
        self._sendMsg(cmd)
        return self._recvMsg()

    def _doRetrieve(self, cmd):
        """Send a command and await its response."""
        self._sendMsg(cmd)
        msg = self._recvMsg()
        result = {}
        error = ""
        while msg.split(" ")[0] == "VALUE":
            key = msg.split(" ")[1]
            flags = int(msg.split(" ")[2]) % 2**32
            length = int(msg.split(" ")[3])
            cas = int(msg.split(" ")[4])
            data = self._recvData(length)
            result[key] = (flags,cas,data)
            msg = self._recvMsg()
        if msg != "END":
            error = msg
        return result, error

    def _doStatsVersion(self, cmd):
        """Send a command and await its response."""
        self._sendMsg(cmd)
        msg = self._recvMsg()
        result = {}
        error = ""
        while msg.split(" ")[0] == "STAT" or \
                msg.split(" ")[0] == "VERSION":
            print "msg:",msg
            kind = msg.split(" ")[0]
            key = msg.split(" ")[1]
            if kind == "VERSION":
                return key, ""
            value = msg.split(" ")[2]
            result[key] = value
            msg = self._recvMsg()
        if msg != "END":
            error = msg
        return result, error

    def _doIncrDecr(self, cmd):
        """Send a command and await its response."""
        self._sendMsg(cmd)
        msg = self._recvMsg()
        try:
            # asci incr/decr doesn't give us the cas
            return (int(msg), 0), ""
        except ValueError:
            return None, msg

    def append(self, key, value, cas=0):
        response = self._doStore("append {0} 0 0 {1} {2}\r\n{3}\r\n".format(key, len(value), cas, value))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def prepend(self, key, value, cas=0):
        response = self._doStore("prepend {0} 0 0 {1} {2}\r\n{3}\r\n".format(key, len(value), cas, value))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def incr(self, key, amt=1, init=0, exp=0):
        """Increment or create the named counter."""
        response, error = self._doIncrDecr("incr {0} {1}\r\n".format(key, amt))
        if error:
            raise MemcachedError(-1, error)
        return response

    def decr(self, key, amt=1, init=0, exp=0):
        """Decrement or create the named counter."""
        response, error = self._doIncrDecr("decr {0} {1}\r\n".format(key, amt))
        if error:
            raise MemcachedError(-1, error)
        return response

    def set(self, key, exp, flags, val):
        """Set a value in the memcached server."""
        response = self._doStore("set {0} {1} {2} {3}\r\n{4}\r\n".format(key, flags, exp, len(val), val))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def add(self, key, exp, flags, val):
        """Add a value in the memcached server iff it doesn't already exist."""
        response = self._doStore("add {0} {1} {2} {3}\r\n{4}\r\n".format(key, flags, exp, len(val), val))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def replace(self, key, exp, flags, val):
        """Replace a value in the memcached server iff it already exists."""
        response = self._doStore("replace {0} {1} {2} {3}\r\n{4}\r\n".format(key, flags, exp, len(val), val))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def get(self, key):
        """Get the value for a given key within the memcached server."""
        response, error = self._doRetrieve("gets {0}\r\n".format(key))
        if error:
            raise MemcachedError(-1, error)
        return response.items()[0][1]

    def getl(self, key, exp=15):
        """Get the value for a given key within the memcached server."""
        response, error = self._doRetrieve("getl {0} {1}\r\n".format(key, exp))
        if error:
            raise MemcachedError(-1, error)
        return response.items()[0][1]

    def cas(self, key, exp, flags, oldVal, val):
        """CAS in a new value for the given key and comparison value."""
        response = self._doStore("cas {0} {1} {2} {3} {4}\r\n{5}\r\n".format(key, flags, exp, len(val), oldVal, val))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def touch(self, key, exp):
        """Touch a key in the memcached server."""
        response = self._doStore("touch {0} {1}\r\n".format(key, exp))
        if response != "STORED":
            raise MemcachedError(-1, response)

    def gat(self, key, exp):
        """Get the value for a given key and touch it within the memcached server."""
        response, error = self._doRetrieve("gat {0} {1}\r\n".format(key, exp))
        if error:
            raise MemcachedError(-1, error)
        return response.items()[0][1]

    def version(self):
        """Get the value for a given key within the memcached server."""
        response, error = self._doStatsVersion("version\r\n")
        if error:
            raise MemcachedError(-1, error)
        return response

    def getMulti(self, keys):
        """Get values for any available keys in the given iterable.
        Returns a dict of matched keys to their values."""

        cmd = "gets"
        for key in keys:
            cmd += " " + key
        cmd += "\r\n"
        response, error = self._doRetrieve(cmd)
        if error:
            raise MemcachedError(-1, error)
        return response

    def stats(self, sub=''):
        """Get stats."""
        if sub:
            sub = " " + sub
        response, error = self._doStatsVersion("stats{0}\r\n".format(sub))
        if error:
            raise MemcachedError(-1, error)
        return response

    def delete(self, key, cas=0):
        """Delete the value for a given key within the memcached server."""
        response = self._doStore("delete {0} {1}\r\n".format(key, cas))
        if response != "DELETED":
            raise MemcachedError(-1, response)

    def flush(self, timebomb=0):
        """Flush all storage in a memcached instance."""
        return self._doCmd(memcacheConstants.CMD_FLUSH, '', '',
            struct.pack(memcacheConstants.FLUSH_PKT_FMT, timebomb))




from membase.api.rest_client import RestConnection, RestHelper
from memcached.helper.data_helper import MemcachedClientHelper
from remote.remote_util import RemoteMachineShellConnection
from mc_bin_client import MemcachedClient, MemcachedError
from membase.api.exception import ServerAlreadyJoinedException
from membase.helper.rebalance_helper import RebalanceHelper
import memcacheConstants

import logger
import testconstants
import time
import Queue
from threading import Thread
import traceback


class ClusterOperationHelper(object):
    #the first ip is taken as the master ip

    # Returns True if cluster successfully finished ther rebalance
    @staticmethod
    def add_and_rebalance(servers, wait_for_rebalance=True):
        log = logger.Logger.get_logger()
        master = servers[0]
        all_nodes_added = True
        rebalanced = True
        rest = RestConnection(master)
        if len(servers) > 1:
            for serverInfo in servers[1:]:
                log.info('adding node : {0}:{1} to the cluster'.format(
                        serverInfo.ip, serverInfo.port))
                otpNode = rest.add_node(master.rest_username, master.rest_password, serverInfo.ip, port=serverInfo.port)
                if otpNode:
                    log.info('added node : {0} to the cluster'.format(otpNode.id))
                else:
                    all_nodes_added = False
                    break
            if all_nodes_added:
                rest.rebalance(otpNodes=[node.id for node in rest.node_statuses()], ejectedNodes=[])
                if wait_for_rebalance:
                    rebalanced &= rest.monitorRebalance()
                else:
                    rebalanced = False
        return all_nodes_added and rebalanced

    # For a clearer API
    @staticmethod
    def remove_and_rebalance(servers, wait_for_rebalance=True):
        return ClusterOperationHelper.cleanup_cluster(
            servers, wait_for_rebalance)

    @staticmethod
    def add_all_nodes_or_assert(master, all_servers, rest_settings, test_case):
        log = logger.Logger.get_logger()
        otpNodes = []
        all_nodes_added = True
        rest = RestConnection(master)
        for serverInfo in all_servers:
            if serverInfo.ip != master.ip:
                log.info('adding node : {0}:{1} to the cluster'.format(
                        serverInfo.ip, serverInfo.port))
                otpNode = rest.add_node(rest_settings.rest_username,
                                        rest_settings.rest_password,
                                        serverInfo.ip)
                if otpNode:
                    log.info('added node : {0} to the cluster'.format(otpNode.id))
                    otpNodes.append(otpNode)
                else:
                    all_nodes_added = False
        if not all_nodes_added:
            if test_case:
                test_case.assertTrue(all_nodes_added,
                                     msg="unable to add all the nodes to the cluster")
            else:
                log.error("unable to add all the nodes to the cluster")
        return otpNodes


    @staticmethod
    def wait_for_ns_servers_or_assert(servers, testcase):
        for server in servers:
            rest = RestConnection(server)
            log = logger.Logger.get_logger()
            log.info("waiting for ns_server @ {0}:{1}".format(server.ip, server.port))
            testcase.assertTrue(RestHelper(rest).is_ns_server_running(),
                                "ns_server is not running in {0}".format(server.ip))

    @staticmethod
    def verify_persistence(servers, test, keys_count=400000, timeout_in_seconds=300):
        log = logger.Logger.get_logger()
        master = servers[0]
        rest = RestConnection(master)
        log.info("Verifying Persistence")
        buckets = rest.get_buckets()
        for bucket in buckets:
        #Load some data
            l_threads = MemcachedClientHelper.create_threads([master], bucket.name,
                                                                     - 1, keys_count, {1024: 0.50, 512: 0.50}, 2, -1,
                                                                     True, True)
            [t.start() for t in l_threads]
            # Do persistence verification
            ready = ClusterOperationHelper.persistence_verification(servers, bucket.name, timeout_in_seconds)
            log.info("Persistence Verification returned ? {0}".format(ready))
            log.info("waiting for persistence threads to finish...")
            for t in l_threads:
                t.aborted = True
            for t in l_threads:
                t.join()
            log.info("persistence thread has finished...")
            test.assertTrue(ready, msg="Cannot verify persistence")


    @staticmethod
    def persistence_verification(servers, bucket, timeout_in_seconds=1260):
        log = logger.Logger.get_logger()
        verification_threads = []
        queue = Queue.Queue()
        rest = RestConnection(servers[0])
        nodes = rest.get_nodes()
        nodes_ip = []
        for node in nodes:
            nodes_ip.append(node.ip)
        for i in range(len(servers)):
            if servers[i].ip in nodes_ip:
                log.info("Server {0}:{1} part of cluster".format(
                        servers[i].ip, servers[i].port))
                rest = RestConnection(servers[i])
                t = Thread(target=ClusterOperationHelper.persistence_verification_per_node,
                           name="verification-thread-{0}".format(servers[i]),
                           args=(rest, bucket, queue, timeout_in_seconds))
                verification_threads.append(t)
        for t in verification_threads:
            t.start()
        for t in verification_threads:
            t.join()
            log.info("thread {0} finished".format(t.name))
        while not queue.empty():
            item = queue.get()
            if item is False:
                return False
        return True

    @staticmethod
    def persistence_verification_per_node(rest, bucket, queue=None, timeout=1260):
        log = logger.Logger.get_logger()
        stat_key = 'ep_flusher_todo'
        start = time.time()
        stats = []
        # Collect stats data points
        while time.time() - start <= timeout:
            _new_stats = rest.get_bucket_stats(bucket)
            if _new_stats and 'ep_flusher_todo' in _new_stats:
                stats.append(_new_stats[stat_key])
                time.sleep(0.5)
            else:
                log.error("unable to obtain stats for bucket : {0}".format(bucket))
        value_90th = ClusterOperationHelper.percentile(stats, 90)
        average = float(sum(stats)) / len(stats)
        log.info("90th percentile value is {0} and average {1}".format(value_90th, average))
        if value_90th == 0 and average == 0:
            queue.put(False)
            return
        queue.put(True)

    @staticmethod
    def percentile(samples, percentile):
        element_idx = int(len(samples) * (percentile / 100.0))
        samples.sort()
        value = samples[element_idx]
        return value

    @staticmethod
    def start_cluster(servers):
        for server in servers:
            shell = RemoteMachineShellConnection(server)
            if shell.is_membase_installed():
                shell.start_membase()
            else:
                shell.start_couchbase()

    @staticmethod
    def stop_cluster(servers):
        for server in servers:
            shell = RemoteMachineShellConnection(server)
            if shell.is_membase_installed():
                shell.stop_membase()
            else:
                shell.stop_couchbase()

    @staticmethod
    def cleanup_cluster(servers, wait_for_rebalance=True):
        log = logger.Logger.get_logger()
        rest = RestConnection(servers[0])
        helper = RestHelper(rest)
        helper.is_ns_server_running(timeout_in_seconds=testconstants.NS_SERVER_TIMEOUT)
        nodes = rest.node_statuses()
        master_id = rest.get_nodes_self().id
        if len(nodes) > 1:
            log.info("rebalancing all nodes in order to remove nodes")
            removed = helper.remove_nodes(knownNodes=[node.id for node in nodes],
                                          ejectedNodes=[node.id for node in nodes if node.id != master_id],
                                          wait_for_rebalance=wait_for_rebalance)
            log.info("removed all the nodes from cluster associated with {0} ? {1}".format(servers[0], removed))

    @staticmethod
    def flushctl_start(servers, username=None, password=None):
        for server in servers:
            c = MemcachedClient(server.ip, 11210)
            if username:
                c.sasl_auth_plain(username, password)
            c.start_persistence()

    @staticmethod
    def flushctl_stop(servers, username=None, password=None):
        for server in servers:
            c = MemcachedClient(server.ip, 11210)
            if username:
                c.sasl_auth_plain(username, password)
            c.stop_persistence()

    @staticmethod
    def flush_os_caches(servers):
        log = logger.Logger.get_logger()
        for server in servers:
            try:
                shell = RemoteMachineShellConnection(server)
                shell.flush_os_caches()
                log.info("Clearing os caches on {0}".format(server))
            except:
                pass

    @staticmethod
    def flushctl_set(master, key, val, bucket='default'):
        rest = RestConnection(master)
        servers = rest.get_nodes()
        for server in servers:
            _server = {"ip": server.ip, "port": master.port,
                       "username": master.rest_username,
                       "password": master.rest_password}
            ClusterOperationHelper.flushctl_set_per_node(_server, key, val, bucket)

    @staticmethod
    def flushctl_set_per_node(server, key, val, bucket='default'):
        log = logger.Logger.get_logger()
        rest = RestConnection(server)
        node = rest.get_nodes_self()
        mc = MemcachedClientHelper.direct_client(server, bucket)
        log.info("Setting flush param on server {0}, {1} to {2} on {3}".format(server, key, val, bucket))
        # Workaround for CBQE-249, ideally this should be node.version
        index_path = node.storage[0].get_index_path()
        if index_path is '':
            # Indicates non 2.0 build
            rv = mc.set_flush_param(key, str(val))
        else:
            type = ClusterOperationHelper._get_engine_param_type(key)
            rv = mc.set_param(key, str(val), type)
        log.info("Setting flush param on server {0}, {1} to {2}, result: {3}".format(server, key, val, rv))
        mc.close()

    @staticmethod
    def _get_engine_param_type(key):
        tap_params = ['tap_keepalive', 'tap_throttle_queue_cap', 'tap_throttle_threshold']
        checkpoint_params = ['chk_max_items', 'chk_period', 'inconsistent_slave_chk', 'keep_closed_chks',
                             'max_checkpoints', 'item_num_based_new_chk']
        flush_params = ['bg_fetch_delay', 'couch_response_timeout', 'exp_pager_stime', 'flushall_enabled',
                        'klog_compactor_queue_cap', 'klog_max_log_size', 'klog_max_entry_ratio',
                        'queue_age_cap', 'max_size', 'max_txn_size', 'mem_high_wat', 'mem_low_wat',
                        'min_data_age', 'timing_log', 'alog_sleep_time']
        if key in tap_params:
            return memcacheConstants.ENGINE_PARAM_TAP
        if key in checkpoint_params:
            return memcacheConstants.ENGINE_PARAM_CHECKPOINT
        if key in flush_params:
            return memcacheConstants.ENGINE_PARAM_FLUSH

    @staticmethod
    def set_expiry_pager_sleep_time(master, bucket, value=30):
        log = logger.Logger.get_logger()
        rest = RestConnection(master)
        servers = rest.get_nodes()
        for server in servers:
            #this is not bucket specific so no need to pass in the bucketname
            log.info("connecting to memcached {0}:{1}".format(server.ip, server.memcached))
            mc = MemcachedClientHelper.direct_client(server, bucket)
            log.info("Set exp_pager_stime flush param on server {0}:{1}".format(server.ip, server.port))
            try:
                mc.set_flush_param("exp_pager_stime", str(value))
                log.info("Set exp_pager_stime flush param on server {0}:{1}".format(server.ip, server.port))
            except Exception as ex:
                traceback.print_exc()
                log.error("Unable to set exp_pager_stime flush param on memcached {0}:{1}".format(server.ip, server.memcached))

    @staticmethod
    def get_mb_stats(servers, key):
        log = logger.Logger.get_logger()
        for server in servers:
            c = MemcachedClient(server.ip, 11210)
            log.info("Get flush param on server {0}, {1}".format(server, key))
            value = c.stats().get(key, None)
            log.info("Get flush param on server {0}, {1}".format(server, value))
            c.close()

    @staticmethod
    def change_erlang_async(servers, original, modified):
        log = logger.Logger.get_logger()
        for server in servers:
            sh = RemoteMachineShellConnection(server)
            command = "sed -i 's/+A {0}/+A {1}/g' /opt/couchbase/bin/membase-server".format(original, modified)
            o, r = sh.execute_command(command)
            sh.log_command_output(o, r)
            msg = "modified erlang +A from {0} to {1} for server {2}"
            log.info(msg.format(original, modified, server.ip))

    @staticmethod
    def begin_rebalance_in(master, servers, timeout=5):
        RebalanceHelper.begin_rebalance_in(master, servers, timeout)

    @staticmethod
    def begin_rebalance_out(master, servers, timeout=5):
        RebalanceHelper.begin_rebalance_out(master, servers, timeout)

    @staticmethod
    def end_rebalance(master):
        RebalanceHelper.end_rebalance(master)

    @staticmethod
    # Returns the otpNode for Orchestrator
    def find_orchestrator(master):
        rest = RestConnection(master)
        command = "node(global:whereis_name(ns_orchestrator))"
        status, content = rest.diag_eval(command)
        # Get rid of single quotes 'ns_1@10.1.3.74'
        content = content.replace("'", '')
        return status, content

    @staticmethod
    def set_vbuckets(master, vbuckets):
        rest = RestConnection(master)
        command = "rpc:eval_everywhere(ns_config, set, [couchbase_num_vbuckets_default, {0}]).".format(vbuckets)
        status, content = rest.diag_eval(command)
        return status, content

    @staticmethod
    def wait_for_completion(rest, task_type='indexer'):
        """Wait for ns_server task (defined by task_type argument) to finish.

        Known tasks:
        -- indexer
        -- view_compaction
        """

        tasks = rest.ns_server_tasks

        time.sleep(30)
        while([task for task in tasks() if task['type'] == task_type]):
            print "Waiting for {0} to finish".format(task_type)
            time.sleep(10)



import json
import random
import logger
import time
import unittest

from TestInput import TestInputSingleton
from membase.api.rest_client import RestConnection, RestHelper
from membase.helper.bucket_helper import BucketOperationHelper
from membase.helper.cluster_helper import ClusterOperationHelper
from membase.helper.rebalance_helper import RebalanceHelper
from memcached.helper.data_helper import MemcachedClientHelper
from memcached.helper.data_helper import MemcachedError
import memcacheConstants


# The SpatialHelper operates on a single bucket over a single RestConnection
# The original testcase needs to be passed in so we can make assertions
class SpatialHelper:
    def __init__(self, testcase, bucket):
        self.testcase = testcase
        self.bucket = bucket
        self.servers = TestInputSingleton.input.servers
        self.input = TestInputSingleton.input
        self.master = self.servers[0]
        self.rest = RestConnection(self.master)
        self.log = logger.Logger.get_logger()

        # A set of indexes that were created with create_spatial_index_fun
        # It's a set, so indexes can be updated without problems
        self._indexes = set([])


    def setup_cluster(self, do_rebalance=True):
        node_ram_ratio = BucketOperationHelper.base_bucket_ratio(self.servers)
        mem_quota = int(self.rest.get_nodes_self().mcdMemoryReserved *
                        node_ram_ratio)
        self.rest.init_cluster(self.master.rest_username,
                               self.master.rest_password)
        self.rest.init_cluster_memoryQuota(self.master.rest_username,
                                      self.master.rest_password,
                                      memoryQuota=mem_quota)
        for server in self.servers:
            ClusterOperationHelper.cleanup_cluster([server])
        ClusterOperationHelper.wait_for_ns_servers_or_assert(
            [self.master], self.testcase)

        if do_rebalance:
            rebalanced = ClusterOperationHelper.add_and_rebalance(
                self.servers)
            self.testcase.assertTrue(rebalanced, "cluster is not rebalanced")

        self._create_default_bucket()


    def cleanup_cluster(self):
        if not "skip_cleanup" in TestInputSingleton.input.test_params:
            # Cleanup all indexes that were create with this helper class
            for name in self._indexes:
                self.rest.delete_spatial(self.bucket, name)
                self.log.info("deleted spatial {0} from bucket {1}"
                              .format(name, self.bucket))
            BucketOperationHelper.delete_all_buckets_or_assert(
                self.servers, self.testcase)
            ClusterOperationHelper.cleanup_cluster(self.servers)
            ClusterOperationHelper.wait_for_ns_servers_or_assert(
                self.servers, self.testcase)


    def create_index_fun(self, name, prefix='', fun=None):
        if fun is None:
            fun = 'function (doc, meta) {if(meta.id.indexOf("' + prefix + \
                '") != -1) { emit(doc.geometry, doc);}}'
        function = self._create_function(name, fun)
        self.rest.create_spatial(self.bucket, name, function)
        self._indexes.add(name)


    # If you insert docs that are already there, they are simply
    # overwritten.
    # extra_values are key value pairs that will be added to the
    # JSON document
    # If `return_docs` is true, it'll return the full docs and not
    # only the keys
    def insert_docs(self, num_of_docs, prefix, extra_values={},
                    wait_for_persistence=True, return_docs=False):
        moxi = MemcachedClientHelper.proxy_client(self.master, self.bucket)
        doc_names = []
        for i in range(0, num_of_docs):
            key = doc_name = "{0}-{1}".format(prefix, i)
            geom = {"type": "Point", "coordinates":
                        [random.randrange(-180, 180),
                         random.randrange(-90, 90)]}
            value = {"name": doc_name, "age": 1000, "geometry": geom}
            value.update(extra_values)
            if not return_docs:
                doc_names.append(doc_name)
            else:
                doc_names.append(value)
            # loop till value is set
            fail_count = 0
            while True:
                try:
                    moxi.set(key, 0, 0, json.dumps(value))
                    break
                except MemcachedError as e:
                    fail_count += 1
                    if (e.status == 133 or e.status == 132) and fail_count < 60:
                        if i == 0:
                            self.log.error("moxi not fully restarted, "
                                           "waiting 5 seconds. error {0}"
                                           .format(e))
                            time.sleep(5)
                        else:
                            self.log.error(e)
                            time.sleep(1)
                    else:
                        raise e
        if wait_for_persistence:
            self.wait_for_persistence()
        self.log.info("inserted {0} json documents".format(num_of_docs))
        return doc_names


    # Returns the keys of the deleted documents
    # If you try to delete a document that doesn't exists, just skip it
    def delete_docs(self, num_of_docs, prefix):
        moxi = MemcachedClientHelper.proxy_client(self.master, self.bucket)
        doc_names = []
        for i in range(0, num_of_docs):
            key = "{0}-{1}".format(prefix, i)
            try:
                moxi.delete(key)
            except MemcachedError as e:
                # Don't care if we try to delete a document that doesn't exist
                if e.status == memcacheConstants.ERR_NOT_FOUND:
                    continue
                else:
                    raise
            doc_names.append(key)

        self.wait_for_persistence(180)

        self.log.info("deleted {0} json documents".format(len(doc_names)))
        return doc_names

    # If `num_expected` is passed in, the number of returned results will
    # be checked. If it doesn't match try again to get the results
    def get_results(self, spatial, limit=None, extra_params={},
                    num_expected=None):
        start = time.time()
        for i in range(0, 4):
            try:
                #full_set=true&connection_timeout=60000&limit=10&skip=0
                params = {"connection_timeout": 60000}
                params.update(extra_params)
                # Make "full_set=true" the default
                if not "full_set" in params:
                    params["full_set"] = True
                results = self.rest.spatial_results(self.bucket, spatial,
                                                    params, limit)
                delta = time.time() - start
                if results:
                    # Keep on trying until we have at least the number
                    # of expected rows
                    if (num_expected is not None) and \
                            (len(results["rows"]) < num_expected):
                        continue

                    self.log.info("spatial returned in {0} seconds"
                                  .format(delta))
                    self.log.info("was able to get spatial results after "
                                  "trying {0} times".format((i + 1)))
                    return results
            except Exception as ex:
                self.log.error("spatial_results not ready yet , try again "
                               "in 5 seconds... , error {0}".format(ex))
                time.sleep(5)

        # Can't get the correct result, fail the test
        self.log.info("num_expected: {0}".format(num_expected))
        self.log.info("num results:  {0}".format(len(results["rows"])))
        self.testcase.fail(
            "unable to get spatial_results for {0} after 4 tries"
            .format(spatial))


    def info(self, spatial):
        return self.rest.spatial_info(self.bucket, spatial)


    # A request to perform the compaction normally returns immediately after
    # it is starting, without waiting until it's completed. This function
    # call keeps blocking until the compaction is done.
    # If you pass in False as a second parameter, it won't block and return
    # immediately
    # Returns True if the compaction succeded within the given timeout
    def compact(self, spatial, block=True, timeout=60):
        status, value = self.rest.spatial_compaction(self.bucket, spatial)
        if not status:
            raise Exception("Compaction returned error.")

        while True:
            status, info = self.info(spatial)
            if not info["spatial_index"]["compact_running"]:
                return True
            elif timeout < 0:
                raise Exception("Compaction timed out.")
            else:
                time.sleep(1)
                timeout -= 1


    def _create_function(self, name, function):
        #if this view already exist then get the rev and embed it here?
        doc = {"language": "javascript"}
        doc["spatial"] = {name: function}
        self.log.info("doc {0}".format(doc))
        return json.dumps(doc)


    #create a bucket if it doesn't exist
    def _create_default_bucket(self):
        helper = RestHelper(self.rest)
        if not helper.bucket_exists(self.bucket):
            node_ram_ratio = BucketOperationHelper.base_bucket_ratio(
                self.servers)
            info = self.rest.get_nodes_self()
            available_ram = int(info.memoryQuota * node_ram_ratio)
            if available_ram < 256:
                available_ram = 256
            self.rest.create_bucket(bucket=self.bucket,
                                    ramQuotaMB=available_ram)
            ready = BucketOperationHelper.wait_for_memcached(self.master,
                                                             self.bucket)
            self.testcase.assertTrue(ready, "wait_for_memcached failed")
        self.testcase.assertTrue(
            helper.bucket_exists(self.bucket),
            "unable to create {0} bucket".format(self.bucket))


    # Return the keys (document ids) of a spatial view response
    def get_keys(self, results):
        keys = []
        if results:
            rows = results["rows"]
            for row in rows:
                keys.append(row["id"].encode("ascii", "ignore"))
            self.log.info("there are {0} keys".format(len(keys)))
        return keys


    # Verify that the built index is correct. Wait until all data got
    # persited on disk
    # If `full_docs` is true, the whole documents, not only the keys
    # will be verified
    # Note that the resultset might be way bigger, we only check if
    # the keys that should be inserted, were really inserted (i.e. that
    # there might be additional keys returned)
    def query_index_for_verification(self, design_name, inserted,
                                     full_docs=False,
                                     wait_for_persistence=True):
        if wait_for_persistence:
            self.wait_for_persistence()
        results = self.get_results(design_name, num_expected=len(inserted))
        result_keys = self.get_keys(results)

        if not full_docs:
            self.verify_result(inserted, result_keys)
        else:
            # The default spatial view function retrurns the full
            # document, hence the values can be used for the
            # meta information as well
            # Only verify things we can get from the value, hence not
            # the revision, nor the bbox. For the ID we use the "name"
            # of the value
            inserted_expanded = []
            for value in inserted:
                inserted_expanded.append(
                    json.dumps({'id': value['name'],
                                'geometry': value['geometry'],
                                'value': value}, sort_keys=True))

            results_collapsed = []
            for row in results['rows']:
                # Delete all top level key-values that are not part of the
                # inserted_expanded list
                del_keys = set(row.keys()) - set(['id', 'geometry', 'value'])
                for key in del_keys:
                    del row[key]
                # Delete all special values inserted by CouchDB or Couchbase
                for key in row['value'].keys():
                    if key.startswith('_') or key.startswith('$'):
                        del row['value'][key]
                results_collapsed.append(json.dumps(row, sort_keys=True))

            diff = set(inserted_expanded) - set(results_collapsed)
            self.testcase.assertEqual(diff, set())


    # Compare the inserted documents with the returned result
    # Both arguments contain a list of document names
    def verify_result(self, inserted, result):
        #not_found = []
        #for key in inserted:
        #    if not key in result:
        #        not_found.append(key)
        not_found = list(set(inserted) - set(result))
        if not_found:
            self._print_keys_not_found(not_found)
            self.testcase.fail("the spatial function did return only {0} "
                               "docs and not {1} as expected"
                               .format(len(result), len(inserted)))


    def _print_keys_not_found(self, keys_not_found, how_many=10):
        how_many = min(how_many, len(keys_not_found))

        for i in range(0, how_many):
            self.log.error("did not find key {0} in the spatial view results"
                           .format(keys_not_found[i]))

    def wait_for_persistence(self, timeout=120):
        RebalanceHelper.wait_for_persistence(self.master, self.bucket, timeout)



#
# Copyright 2008 Orbitz WorldWide
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import struct
import time
import operator
import itertools
import re

from whisper.exceptions import InvalidConfiguration, \
    InvalidAggregationMethod, InvalidTimeInterval, TimestampNotCovered, \
    CorruptWhisperFile

try:
    import fcntl
    CAN_LOCK = True
except ImportError:
    CAN_LOCK = False


class Whisper(object):

    """This module is an implementation of the Whisper database API.
    Here is the basic layout of a whisper data file:

    file = header, data
        header = metadata, archive_info
            metadata = aggregation_type, max_retention, x_files_factor, archive_count
            archiveInfo = offset, seconds_per_point, points
        data = archive
            archive = point
                point = timestamp, value
    """

    CAN_LOCK = CAN_LOCK
    LOCK = False
    CACHE_HEADERS = False
    AUTOFLUSH = False

    LONG_FORMAT = "!L"
    LONG_SIZE = struct.calcsize(LONG_FORMAT)
    FLOAT_FORMAT = "!f"
    FLOAT_SIZE = struct.calcsize(FLOAT_FORMAT)
    VALUE_FORMAT = "!d"
    VALUE_SIZE = struct.calcsize(VALUE_FORMAT)
    POINT_FORMAT = "!Ld"
    POINT_SIZE = struct.calcsize(POINT_FORMAT)
    METADATA_FORMAT = "!2LfL"
    METADATA_SIZE = struct.calcsize(METADATA_FORMAT)
    ARCHIVE_INFO_FORMAT = "!3L"
    ARCHIVE_INFO_SIZE = struct.calcsize(ARCHIVE_INFO_FORMAT)

    AGGREGATION_TYPE_TO_METHOD = {1: 'average',
                                  2: 'sum',
                                  3: 'last',
                                  4: 'max',
                                  5: 'min'}
    AGGREGATION_METHOD_TO_TYPE = \
        dict((v, k) for k, v in AGGREGATION_TYPE_TO_METHOD.iteritems())
    AGGREGATION_METHODS = AGGREGATION_TYPE_TO_METHOD.values()

    UNIT_MULTIPLIERS = {'seconds': 1,
                        'minutes': 60,
                        'hours': 3600,
                        'days': 86400,
                        'weeks': 86400 * 7,
                        'years': 86400 * 365}

    _header_cache = {}

    @staticmethod
    def get_unit_string(s):
        if 'seconds'.startswith(s):
            return 'seconds'
        if 'minutes'.startswith(s):
            return 'minutes'
        if 'hours'.startswith(s):
            return 'hours'
        if 'days'.startswith(s):
            return 'days'
        if 'weeks'.startswith(s):
            return 'weeks'
        if 'years'.startswith(s):
            return 'years'
        raise ValueError("Invalid unit '%s'" % s)

    @staticmethod
    def parse_retention_def(retention_def):
        (precision, points) = retention_def.strip().split(':')

        if precision.isdigit():
            precision = int(precision) * \
                Whisper.UNIT_MULTIPLIERS[Whisper.get_unit_string('s')]
        else:
            precision_re = re.compile(r'^(\d+)([a-z]+)$')
            match = precision_re.match(precision)
            if match:
                precision = int(match.group(1)) * \
                    Whisper.UNIT_MULTIPLIERS[Whisper.get_unit_string(match.group(2))]
            else:
                raise ValueError(
                    "Invalid precision specification '%s'" % precision)

        if points.isdigit():
            points = int(points)
        else:
            points_re = re.compile(r'^(\d+)([a-z]+)$')
            match = points_re.match(points)
            if match:
                points = int(match.group(1)) * \
                    Whisper.UNIT_MULTIPLIERS[Whisper.get_unit_string(match.group(2))] / precision
            else:
                raise ValueError("Invalid retention specification '%s'" %
                                 points)

        return (precision, points)

    @staticmethod
    def _read_header(fh):
        info = Whisper._header_cache.get(fh.name)
        if info:
            return info

        original_offset = fh.tell()
        fh.seek(0)
        packed_metadata = fh.read(Whisper.METADATA_SIZE)

        try:
            (aggregation_type, max_retention, xff, archive_count) = \
                struct.unpack(Whisper.METADATA_FORMAT, packed_metadata)
        except:
            raise CorruptWhisperFile("Unable to read header", fh.name)

        archives = []

        for i in xrange(archive_count):
            packed_archive_info = fh.read(Whisper.ARCHIVE_INFO_SIZE)
            try:
                (offset, seconds_per_point, points) = \
                    struct.unpack(Whisper.ARCHIVE_INFO_FORMAT,
                                  packed_archive_info)
            except:
                raise CorruptWhisperFile("Unable to read archive%d metadata" %
                                         i, fh.name)

            archive_info = {'offset': offset,
                            'seconds_per_point': seconds_per_point,
                            'points': points,
                            'retention': seconds_per_point * points,
                            'size': points * Whisper.POINT_SIZE}
            archives.append(archive_info)

        fh.seek(original_offset)
        aggregation_method = \
            Whisper.AGGREGATION_TYPE_TO_METHOD.get(aggregation_type,
                                                   'average')
        info = {'aggregation_method': aggregation_method,
                'max_retention': max_retention,
                'x_files_factor': xff,
                'archives': archives}
        if Whisper.CACHE_HEADERS:
            Whisper._header_cache[fh.name] = info

        return info

    @staticmethod
    def set_aggregation_method(path, aggregation_method):
        """
        path -- a string
        aggregation_method -- specifies the method to use when propogating data
        (see ``whisper.AGGREGATION_METHODS``)
        """
        with open(path, 'r+b') as fh:
            if Whisper.LOCK:
                fcntl.flock(fh.fileno(), fcntl.LOCK_EX)

            packed_metadata = fh.read(Whisper.METADATA_SIZE)

            try:
                (aggregation_type, max_retention, xff, archive_count) = \
                    struct.unpack(Whisper.METADATA_FORMAT, packed_metadata)
            except:
                raise CorruptWhisperFile("Unable to read header", fh.name)

            try:
                new_aggregation_type = \
                    Whisper.AGGREGATION_METHOD_TO_TYPE[aggregation_method]
                new_aggregation_type = struct.pack(Whisper.LONG_FORMAT,
                                                   new_aggregation_type)
            except KeyError:
                err_message = "Unrecognized aggregation method: %s" % \
                    aggregation_method
                raise InvalidAggregationMethod(err_message)

            fh.seek(0)
            fh.write(new_aggregation_type)

            if Whisper.AUTOFLUSH:
                fh.flush()
                os.fsync(fh.fileno())

            if Whisper.CACHE_HEADERS and fh.name in Whisper._header_cache:
                del Whisper._header_cache[fh.name]

            return Whisper.AGGREGATION_TYPE_TO_METHOD.get(aggregation_type,
                                                          'average')

    @staticmethod
    def validate_archive_list(archive_list):
        """Validates an archive_list.
        An ArchiveList must:
        1. Have at least one archive config. Example: (60, 86400)
        2. No archive may be a duplicate of another.
        3. Higher precision archives' precision must evenly divide all lower
        precision archives' precision.
        4. Lower precision archives must cover larger time intervals than
        higher precision archives.
        5. Each archive must have at least enough points to consolidate to the
        next archive

        Returns True or False
        """

        if not archive_list:
            raise InvalidConfiguration("You must specify at least one archive"
                                       "configuration!")

        archive_list.sort(
            key=lambda a: a[0])  # sort by precision (seconds_per_point)

        for i, archive in enumerate(archive_list):
            if i == len(archive_list) - 1:
                break

            next_archive = archive_list[i + 1]
            if not archive[0] < next_archive[0]:
                err_message = "A Whisper database may not configured having " \
                              "two archives with the same precision " \
                              "(archive%d: %s, archive%d: %s)" % \
                              (i, archive, i + 1, next_archive)
                raise InvalidConfiguration(err_message)

            if next_archive[0] % archive[0] != 0:
                err_message = "Higher precision archives must evenly divide " \
                              "all lower precision archives precision " \
                              "(archive%d: %s, archive%d: %s)" % \
                              (i, archive[0], i + 1, next_archive[0])
                raise InvalidConfiguration(err_message)

            retention = archive[0] * archive[1]
            next_retention = next_archive[0] * next_archive[1]

            if not next_retention > retention:
                err_message = "Lower precision archives must cover larger " \
                              "time intervals than higher precision archives" \
                              "(archive%d: %s seconds, archive%d: %s " \
                              "seconds)" % \
                              (i, archive[1], i + 1, next_archive[1])
                raise InvalidConfiguration(err_message)

            archive_points = archive[1]
            points_per_consolidation = next_archive[0] / archive[0]
            if not archive_points >= points_per_consolidation:
                err_message = "Each archive must have at least enough points" \
                              "to consolidate to the next archive " \
                              "(archive%d consolidates %d of archive%d's " \
                              "points but it has only %d total points)" % \
                              (i + 1, points_per_consolidation, i,
                              archive_points)
                raise InvalidConfiguration(err_message)

    @staticmethod
    def create(path, archive_list, x_files_factor=None,
               aggregation_method=None, sparse=False):
        """
        path -- a string

        archive_list -- a list of archives, each of which is of the form
        (seconds_per_point,number_of_points).

        x_files_factor -- specifies the fraction of data points in a
        propagation interval that must have known values for a propagation to
        occur.

        aggregation_method -- specifies the function to use when propogating
        data (see ``whisper.AGGREGATION_METHODS``)
        """
        # Set default params
        if x_files_factor is None:
            x_files_factor = 0.5
        if aggregation_method is None:
            aggregation_method = 'average'

        #Validate archive configurations...
        Whisper.validate_archive_list(archive_list)

        #Looks good, now we create the file and write the header
        if os.path.exists(path):
            raise InvalidConfiguration("File %s already exists!" % path)

        with open(path, 'wb') as fh:
            if Whisper.LOCK:
                fcntl.flock(fh.fileno(), fcntl.LOCK_EX)

            aggregation_type = \
                Whisper.AGGREGATION_METHOD_TO_TYPE.get(aggregation_method, 1)
            aggregation_type = struct.pack(Whisper.LONG_FORMAT,
                                           aggregation_type)
            oldest = max(seconds_per_point * points
                         for seconds_per_point, points in archive_list)
            max_retention = struct.pack(Whisper.LONG_FORMAT, oldest)
            x_files_factor = struct.pack(
                Whisper.FLOAT_FORMAT, float(x_files_factor))
            archive_count = struct.pack(Whisper.LONG_FORMAT, len(archive_list))
            packed_metadata = aggregation_type + max_retention + \
                x_files_factor + archive_count
            fh.write(packed_metadata)
            header_size = Whisper.METADATA_SIZE + \
                (Whisper.ARCHIVE_INFO_SIZE * len(archive_list))
            archive_offset_pointer = header_size

            for seconds_per_point, points in archive_list:
                archive_info = struct.pack(Whisper.ARCHIVE_INFO_FORMAT,
                                           archive_offset_pointer,
                                           seconds_per_point, points)
                fh.write(archive_info)
                archive_offset_pointer += (points * Whisper.POINT_SIZE)

            if sparse:
                fh.seek(archive_offset_pointer - header_size - 1)
                fh.write("\0")
            else:
                # If not creating the file sparsely, then fill the rest of the
                # file
                # with zeroes.
                remaining = archive_offset_pointer - header_size
                chunksize = 16384
                zeroes = '\x00' * chunksize
                while remaining > chunksize:
                    fh.write(zeroes)
                    remaining -= chunksize
                fh.write(zeroes[:remaining])

            if Whisper.AUTOFLUSH:
                fh.flush()
                os.fsync(fh.fileno())

    @staticmethod
    def aggregate(aggregation_method, known_values):
        if aggregation_method == 'average':
            return float(sum(known_values)) / float(len(known_values))
        elif aggregation_method == 'sum':
            return float(sum(known_values))
        elif aggregation_method == 'last':
            return known_values[len(known_values) - 1]
        elif aggregation_method == 'max':
            return max(known_values)
        elif aggregation_method == 'min':
            return min(known_values)
        else:
            err_message = "Unrecognized aggregation method %s" % \
                aggregation_method
            raise InvalidAggregationMethod(err_message)

    @staticmethod
    def _propagate(fh, header, timestamp, higher, lower):
        aggregation_method = header['aggregation_method']
        xff = header['x_files_factor']

        lower_interval_start = timestamp - (timestamp % lower[
            'seconds_per_point'])
        lower_interval_end = lower_interval_start + lower['seconds_per_point']

        fh.seek(higher['offset'])
        packed_point = fh.read(Whisper.POINT_SIZE)
        (higher_base_interval, higher_base_value) = \
            struct.unpack(Whisper.POINT_FORMAT, packed_point)

        if higher_base_interval == 0:
            higher_first_offset = higher['offset']
        else:
            time_distance = lower_interval_start - higher_base_interval
            point_distance = time_distance / higher['seconds_per_point']
            byte_distance = point_distance * Whisper.POINT_SIZE
            higher_first_offset = higher['offset'] + \
                byte_distance % higher['size']

        higher_points = lower['seconds_per_point'] / \
            higher['seconds_per_point']
        higher_size = higher_points * Whisper.POINT_SIZE
        relative_first_offset = higher_first_offset - higher['offset']
        relative_last_offset = (
            relative_first_offset + higher_size) % higher['size']
        higher_last_offset = relative_last_offset + higher['offset']
        fh.seek(higher_first_offset)

        if higher_first_offset < higher_last_offset:
            # we don't wrap the archive
            series_string = fh.read(higher_last_offset - higher_first_offset)
        else:  # We do wrap the archive
            higher_end = higher['offset'] + higher['size']
            series_string = fh.read(higher_end - higher_first_offset)
            fh.seek(higher['offset'])
            series_string += fh.read(higher_last_offset - higher['offset'])

        #Now we unpack the series data we just read
        byte_order, point_types = Whisper.POINT_FORMAT[
            0], Whisper.POINT_FORMAT[1:]
        points = len(series_string) / Whisper.POINT_SIZE
        series_format = byte_order + (point_types * points)
        unpacked_series = struct.unpack(series_format, series_string)

        #And finally we construct a list of values
        neighbor_values = [None] * points
        current_interval = lower_interval_start
        step = higher['seconds_per_point']

        for i in xrange(0, len(unpacked_series), 2):
            point_time = unpacked_series[i]
            if point_time == current_interval:
                neighbor_values[i / 2] = unpacked_series[i + 1]
            current_interval += step

        # Propagate aggregate_value to propagate from neighbor_values if we
        # have enough known points
        known_values = [v for v in neighbor_values if v is not None]
        if not known_values:
            return False

        known_percent = float(len(known_values)) / float(len(neighbor_values))
        if known_percent >= xff:  # we have enough data to propagate a value!
            aggregate_value = Whisper.aggregate(aggregation_method,
                                                known_values)
            my_packed_point = struct.pack(Whisper.POINT_FORMAT,
                                          lower_interval_start,
                                          aggregate_value)
            fh.seek(lower['offset'])
            packed_point = fh.read(Whisper.POINT_SIZE)
            (lower_base_interval, lower_base_value) = \
                struct.unpack(Whisper.POINT_FORMAT, packed_point)

            # First propagated update to this lower archive
            if lower_base_interval == 0:
                fh.seek(lower['offset'])
                fh.write(my_packed_point)
            else:  # Not our first propagated update to this lower archive
                time_distance = lower_interval_start - lower_base_interval
                point_distance = time_distance / lower['seconds_per_point']
                byte_distance = point_distance * Whisper.POINT_SIZE
                lower_offset = lower['offset'] + \
                    (byte_distance % lower['size'])
                fh.seek(lower_offset)
                fh.write(my_packed_point)

            return True

        else:
            return False

    @staticmethod
    def update(path, value, timestamp=None):
        """
        path -- a string
        value -- a float
        timestamp -- either an int or float
        """
        value = float(value)
        with open(path, 'r+b') as fh:
            return Whisper.file_update(fh, value, timestamp)

    @staticmethod
    def file_update(fh, value, timestamp):
        if Whisper.LOCK:
            fcntl.flock(fh.fileno(), fcntl.LOCK_EX)

        header = Whisper._read_header(fh)
        now = int(time.time())
        if timestamp is None:
            timestamp = now

        timestamp = int(timestamp)
        diff = now - timestamp
        if not ((diff < header['max_retention']) and diff >= 0):
            err_message = "Timestamp not covered by any archives in this " \
                "database."
            raise TimestampNotCovered(err_message)

        # Find the highest-precision archive that covers timestamp
        for i, archive in enumerate(header['archives']):
            if archive['retention'] < diff:
                continue
            #We'll pass on the update to these lower precision archives later
            lower_archives = header['archives'][i + 1:]
            break

        #First we update the highest-precision archive
        my_interval = timestamp - (timestamp % archive['seconds_per_point'])
        my_packed_point = struct.pack(Whisper.POINT_FORMAT, my_interval, value)
        fh.seek(archive['offset'])
        packed_point = fh.read(Whisper.POINT_SIZE)
        (base_interval, base_value) = struct.unpack(Whisper.
                                                    POINT_FORMAT, packed_point)

        if base_interval == 0:  # This file's first update
            fh.seek(archive['offset'])
            fh.write(my_packed_point)
            base_interval, base_value = my_interval, value
        else:  # Not our first update
            time_distance = my_interval - base_interval
            point_distance = time_distance / archive['seconds_per_point']
            byte_distance = point_distance * Whisper.POINT_SIZE
            my_offset = archive['offset'] + (byte_distance % archive['size'])
            fh.seek(my_offset)
            fh.write(my_packed_point)

        #Now we propagate the update to lower-precision archives
        higher = archive
        for lower in lower_archives:
            if not Whisper._propagate(fh, header, my_interval, higher, lower):
                break
            higher = lower

        if Whisper.AUTOFLUSH:
            fh.flush()
            os.fsync(fh.fileno())

        fh.close()

    @staticmethod
    def update_many(path, points):
        """
        path -- a string
        points -- a list of (timestamp, value) points
        """
        if not points:
            return
        points = [(int(t), float(v)) for (t, v) in points]
        points.sort(key=lambda p: p[0], reverse=True)
            #order points by timestamp, newest first
        with open(path, 'r+b') as fh:
            return Whisper.file_update_many(fh, points)

    @staticmethod
    def file_update_many(fh, points):
        if Whisper.LOCK:
            fcntl.flock(fh.fileno(), fcntl.LOCK_EX)

        header = Whisper._read_header(fh)
        now = int(time.time())
        archives = iter(header['archives'])
        current_archive = archives.next()
        current_points = []

        for point in points:
            age = now - point[0]

            # we can't fit any more points in this archive
            while current_archive['retention'] < age:
                # commit all the points we've found that it can fit
                if current_points:
                    # put points in chronological order
                    current_points.reverse()
                    Whisper._archive_update_many(
                        fh, header, current_archive, current_points)
                    current_points = []
                try:
                    current_archive = archives.next()
                except StopIteration:
                    current_archive = None
                    break

            if not current_archive:
                break  # drop remaining points that don't fit in the database

            current_points.append(point)
        # don't forget to commit after we've checked all the archives
        if current_archive and current_points:
            current_points.reverse()
            Whisper._archive_update_many(
                fh, header, current_archive, current_points)

        if Whisper.AUTOFLUSH:
            fh.flush()
            os.fsync(fh.fileno())

        fh.close()

    @staticmethod
    def _archive_update_many(fh, header, archive, points):
        step = archive['seconds_per_point']
        aligned_points = [(timestamp - (timestamp % step), value)
                          for (timestamp, value) in points]
        #Create a packed string for each contiguous sequence of points
        packed_strings = []
        previous_interval = None
        current_string = ""
        for (interval, value) in aligned_points:
            if interval == previous_interval:
                continue
            if (not previous_interval) or \
                    (interval == previous_interval + step):
                current_string += struct.pack(Whisper.POINT_FORMAT,
                                              interval, value)
                previous_interval = interval
            else:
                number_of_points = len(current_string) / Whisper.POINT_SIZE
                start_interval = previous_interval - \
                    step * (number_of_points - 1)
                packed_strings.append((start_interval, current_string))
                current_string = struct.pack(Whisper.POINT_FORMAT, interval,
                                             value)
                previous_interval = interval
        if current_string:
            number_of_points = len(current_string) / Whisper.POINT_SIZE
            start_interval = previous_interval - step * (number_of_points - 1)
            packed_strings.append((start_interval, current_string))

        #Read base point and determine where our writes will start
        fh.seek(archive['offset'])
        packed_base_point = fh.read(Whisper.POINT_SIZE)
        (base_interval, base_value) = struct.unpack(Whisper.POINT_FORMAT,
                                                    packed_base_point)
        if base_interval == 0:  # This file's first update
            #use our first string as the base, so we start at the start
            base_interval = packed_strings[0][0]

        # Write all of our packed strings in locations determined by the
        # base_interval
        for (interval, packed_string) in packed_strings:
            time_distance = interval - base_interval
            point_distance = time_distance / step
            byte_distance = point_distance * Whisper.POINT_SIZE
            my_offset = archive['offset'] + (byte_distance % archive['size'])
            fh.seek(my_offset)
            archive_end = archive['offset'] + archive['size']
            bytes_beyond = (my_offset + len(packed_string)) - archive_end

            if bytes_beyond > 0:
                fh.write(packed_string[:-bytes_beyond])
                message = "archive_end=%d fh.tell=%d bytes_beyond=%d " \
                    "len(packed_string)=%d" % (archive_end, fh.tell(),
                                               bytes_beyond,
                                               len(packed_string))
                assert fh.tell() == archive_end, message
                fh.seek(archive['offset'])
                # safe because it can't exceed the archive (retention checking
                # logic above)
                fh.write(packed_string[-bytes_beyond:])
            else:
                fh.write(packed_string)

        #Now we propagate the updates to lower-precision archives
        higher = archive
        lower_archives = [arc for arc in header['archives'] if arc[
            'seconds_per_point'] > archive['seconds_per_point']]

        for lower in lower_archives:
            fit = lambda i: i - (i % lower['seconds_per_point'])
            lower_intervals = [fit(p[0]) for p in aligned_points]
            unique_lower_intervals = set(lower_intervals)
            propagate_further = False
            for interval in unique_lower_intervals:
                if Whisper._propagate(fh, header, interval, higher, lower):
                    propagate_further = True

            if not propagate_further:
                break
            higher = lower

    @staticmethod
    def info(path):
        with open(path, 'rb') as fh:
            return Whisper._read_header(fh)

    @staticmethod
    def fetch(path, from_time, until_time=None):
        """
        path -- a string
        from_time -- an epoch time
        until_time -- also an epoch time, but defaults to now
        """
        with open(path, 'rb') as fh:
            return Whisper.file_fetch(fh, from_time, until_time)

    @staticmethod
    def file_fetch(fh, from_time, until_time):
        header = Whisper._read_header(fh)
        now = int(time.time())
        if until_time is None:
            until_time = now
        from_time = int(from_time)
        until_time = int(until_time)

        oldest_time = now - header['max_retention']
        if from_time < oldest_time:
            from_time = oldest_time

        if not (from_time < until_time):
            raise InvalidTimeInterval("Invalid time interval")
        if until_time > now:
            until_time = now
        if until_time < from_time:
            until_time = now

        diff = now - from_time
        for archive in header['archives']:
            if archive['retention'] >= diff:
                break

        from_interval = int(from_time - (from_time % archive[
            'seconds_per_point'])) + archive['seconds_per_point']
        until_interval = int(until_time -
                             until_time % archive['seconds_per_point'])
        until_interval += archive['seconds_per_point']
        fh.seek(archive['offset'])
        packed_point = fh.read(Whisper.POINT_SIZE)
        (base_interval, base_value) = struct.unpack(Whisper.POINT_FORMAT,
                                                    packed_point)

        if base_interval == 0:
            step = archive['seconds_per_point']
            points = (until_interval - from_interval) / step
            time_info = (from_interval, until_interval, step)
            value_list = [None] * points
            return (time_info, value_list)

        #Determine from_offset
        time_distance = from_interval - base_interval
        point_distance = time_distance / archive['seconds_per_point']
        byte_distance = point_distance * Whisper.POINT_SIZE
        from_offset = archive['offset'] + (byte_distance % archive['size'])

        #Determine until_offset
        time_distance = until_interval - base_interval
        point_distance = time_distance / archive['seconds_per_point']
        byte_distance = point_distance * Whisper.POINT_SIZE
        until_offset = archive['offset'] + (byte_distance % archive['size'])

        #Read all the points in the interval
        fh.seek(from_offset)
        if from_offset < until_offset:  # If we don't wrap around the archive
            series_string = fh.read(until_offset - from_offset)
        else:  # We do wrap around the archive, so we need two reads
            archive_end = archive['offset'] + archive['size']
            series_string = fh.read(archive_end - from_offset)
            fh.seek(archive['offset'])
            series_string += fh.read(until_offset - archive['offset'])

        # Now we unpack the series data we just read (anything faster than
        # unpack?)
        byte_order, point_types = Whisper.POINT_FORMAT[
            0], Whisper.POINT_FORMAT[1:]
        points = len(series_string) / Whisper.POINT_SIZE
        series_format = byte_order + (point_types * points)
        unpacked_series = struct.unpack(series_format, series_string)

        #And finally we construct a list of values (optimize this!)
        value_list = [None] * points  # pre-allocate entire list for speed
        current_interval = from_interval
        step = archive['seconds_per_point']

        for i in xrange(0, len(unpacked_series), 2):
            point_time = unpacked_series[i]
            if point_time == current_interval:
                point_value = unpacked_series[i + 1]
                # in-place reassignment is faster than append()
                value_list[i / 2] = point_value
            current_interval += step

        fh.close()
        time_info = (from_interval, until_interval, step)
        return (time_info, value_list)

    @staticmethod
    def merge(path_from, path_to, step=1 << 12):
        header_from = Whisper.info(path_from)

        archives = header_from['archives']
        archives.sort(key=operator.itemgetter('retention'), reverse=True)

        # Start from max_retention of the oldest file, and skip forward at max
        # 'step'
        # points at a time.
        from_time = int(time.time()) - header_from['max_retention']
        for archive in archives:
            points_remaining = archive['points']
            while points_remaining:
                points_to_read = step
                if points_remaining < step:
                    points_to_read = points_remaining
                points_remaining -= points_to_read
                until_time = from_time + (points_to_read *
                                          archive['seconds_per_point'])
                (time_info, values) = Whisper.fetch(path_from, from_time,
                                                    until_time)
                (start, end, archive_step) = time_info
                points_to_write = list(itertools.ifilter(
                                       lambda points: points[1] is not None,
                                       itertools.izip(xrange(start, end,
                                                             archive_step),
                                                      values)))
                # order points by timestamp,newest first
                points_to_write.sort(key=lambda p: p[0],
                                     reverse=True)
                Whisper.update_many(path_to, points_to_write)
                from_time = until_time



#!/usr/bin/env python

import json
import logging
import os
import re
import sys

"""Helper functions which can recursively traverse or visit couchbase
   REST / management data, driven by metadata.  Users can change
   behavior by passing in different visitor callback functions."""

def visit_dict(root, parents, data, meta, coll, level=0,
               up_key=None, up_data=None, up_coll=None):
    """Invoked when data is a dict."""
    if not isinstance(data, dict):
        log(data, " is not a dict")
        return
    next_level = level + 1
    for key in sorted(data.keys(), cmp=lambda x, y: cmp(hyphen_last(x),
                                                        hyphen_last(y))):
        val = data[key]
        meta_val = meta.get(key, None)
        if meta_val is None:
            log("warning: missing metadata entry at: %s; key: %s" %
                (parents, key))
            continue
        meta_inf = meta.get("-" + key, None)
        visit_entry(root, parents, data, meta, coll,
                    key, val, meta_val, meta_inf, level=next_level)

def visit_list(root, parents, data, meta, coll, level=0,
               up_key=None, up_data=None, up_coll=None):
    """Invoked when data is a list."""
    if not isinstance(data, list):
        log(data, " is not a list")
        return
    next_level = level + 1
    if len(meta) <= 0:
        log("warning: missing list metadata entry at: %s" % (parents))
        return
    meta_val = meta[0] # Use only the first item for metadata.
    for idx, val in enumerate(data):
        visit_entry(root, parents, data, meta, coll,
                    idx, val, meta_val, None, level=next_level)

"""Callbacks when visiting collections of different types.
"""
VISIT_COLLECTION_FUNCS = {"<type 'dict'>": visit_dict,
                          "<type 'list'>": visit_list}

def visit_url(context, path):
    """Recursively visits the JSON response from a URL, driven by metadata
       to follow links and process data."""
    root = dict(context) # Makes a copy.
    root["run"] = {}
    root["run"]["data"] = root["retrieve_funcs"]["retrieve_data"](context, path)
    root["run"]["meta"] = root["retrieve_funcs"]["retrieve_meta"](context, path)
    root["run"]["coll"] = type(root["run"]["meta"])() # Collection/hiearchy of slow data.
    root["run"]["tot_fast"] = 0
    root["run"]["tot_slow"] = 0
    func = root["collection_funcs"][str(type(root["run"]["meta"]))]
    func(root, [],
         root["run"]["data"],
         root["run"]["meta"],
         root["run"]["coll"])
    return root

EMIT_NONE = 0x00
EMIT_SLOW = 0x01
EMIT_FAST = 0x02

def visit_entry_default(root, parents, data, meta, coll,
                        key, val, meta_val, meta_inf, level=0,
                        emit_kind=None):
    """Records a scalar entry into the either the slow-changing or
       fast-changing time-series DB, depending on the emit_kind param.
       Recurses via visit_dict() or visit_list() to handle non-scalar entry."""
    path = parents + [str(key)]
    prefix = "  " * level

    t = type(val)
    if t != type(meta_val):
        if type(meta_val) == float or type(meta_val) == int:
            val = type(meta_val)(val)
        else:
            log("warning: unexpected type: %s; expected %s; at: %s" %
                (t, type(meta_val), path))
            return

    if t == str or t == unicode: # Scalar string entry.
        if emit_kind is None:
            emit_kind = EMIT_SLOW
        if emit_kind & EMIT_SLOW:
            debug(prefix, key, '=', '"%s"' % val)
            root["store"]["slow"](root, parents, data, meta, coll,
                                  key, val, meta_val, meta_inf, level)

            # Handle follow metadata, when val is URL/URI,
            # by pushing work onto the to-be-visited queue.
            if meta_inf and meta_inf.get('follow', None):
                context = dict(root) # Copy the context for recursion.
                del context["run"]   # But, not the current run info.
                root["queue"].append((context, val))

    elif t == float or t == int: # Scalar numeric entry.
        if emit_kind is None:
            emit_kind = EMIT_FAST
        if emit_kind & EMIT_FAST:
            units = ''
            if meta_inf:
                units = meta_inf.get("units", '')
            debug(prefix, "ns-" + '-'.join(path), '=', val, units)
            root["store"]["fast"](root, parents, data, meta, coll,
                                  key, val, meta_val, meta_inf, level)
        if emit_kind & EMIT_SLOW:
            debug(prefix, key, '=', val)
            root["store"]["slow"](root, parents, data, meta, coll,
                                  key, val, meta_val, meta_inf, level)

    elif t == bool: # Scalar boolean entry.
        if emit_kind is None:
            emit_kind = EMIT_SLOW
        if emit_kind & EMIT_FAST:
            debug(prefix, "ns-" + '-'.join(path), '=', int(val))
            root["store"]["fast"](root, parents, data, meta, coll,
                                  key, val, meta_val, meta_inf, level)
            root["run"]["tot_fast"] += 1
        if emit_kind & EMIT_SLOW:
            debug(prefix, key, '=', val)
            root["store"]["slow"](root, parents, data, meta, coll,
                                  key, val, meta_val, meta_inf, level)

    elif t == dict or t == list: # Non-scalar entry.
        child_coll = t()
        debug(prefix, key, "= ...")
        if type(coll) == dict:
            coll[key] = child_coll
        else:
            coll.append(child_coll)
        func = root["collection_funcs"][str(t)]
        func(root, path, val, meta_val, child_coll, level=level,
             up_key=key, up_data=data, up_coll=coll)
    else:
        log("warning: unhandled type for val: %s; in key: %s" % (val, key))

def visit_entry_fast(root, parents, data, meta, coll,
                     key, val, meta_val, meta_inf, level=0):
    """Records a scalar entry into the fast-changing time-series DB."""
    return visit_entry_default(root, parents, data, meta, coll,
                               key, val, meta_val, meta_inf, level=level,
                               emit_kind=EMIT_FAST)

def visit_entry_slow(root, parents, data, meta, coll,
                     key, val, meta_val, meta_inf, level=0):
    """Records a scalar entry into the slow-changing time-series DB."""
    return visit_entry_default(root, parents, data, meta, coll,
                               key, val, meta_val, meta_inf, level=level,
                               emit_kind=EMIT_SLOW)

def visit_entry_int(root, parents, data, meta, coll,
                    key, val, meta_val, meta_inf, level=0):
    """Parse the val as an int, and then use default processing."""
    return visit_entry_default(root, parents, data, meta, coll,
                               key, int(val), int(meta_val), meta_inf, level=level)

def visit_entry_strip(root, parents, data, meta, coll,
                      key, val, meta_val, meta_inf, level=0):
    """This visit_entry_func strips the entry from results."""
    return

def visit_entry_collect_mc_stats(root, parents, data, meta, coll,
                                 key, val, meta_val, meta_inf, level=0):
    """A different implementation could collects memcached stats from the
       val, which should be an array of "HOST:PORT", like
       ["couchbase-01:11210, couchbase-02:11211"].  The root and
       parents path should have bucket and SASL auth info.
       Use the main(entry_funcs) parameter to specify your own implementation."""
    debug("  " * level, "MC-STATS", val)

def visit_entry_collect_proxy_stats(root, parents, data, meta, coll,
                                    key, val, meta_val, meta_inf, level=0):
    """A different implementation could collects proxy stats from the
       val, which should be a port number (like 11211).  The hostname
       should be amongst the ancestors."""
    debug("  " * level, "PROXY-STATS", val)

"""Callbacks when visiting scalar values, driven by 'visit' metadata.
"""
VISIT_ENTRY_FUNCS = {"default": visit_entry_default,
                     "fast": visit_entry_fast,
                     "slow": visit_entry_slow,
                     "int": visit_entry_int,
                     "strip": visit_entry_strip,
                     "collect_mc_stats": visit_entry_collect_mc_stats,
                     "collect_proxy_stats": visit_entry_collect_proxy_stats}

def visit_entry(root, parents, data, meta, coll,
                key, val, meta_val, meta_inf, level=0):
    """Invokes the right visit_entry_func on an entry."""
    if (type(key) == str or type(key) == unicode) and key.startswith('-'):
        if root.get("strip_meta", True):
            return

    if meta_inf:
        visit = meta_inf.get("visit", "default")
    else:
        visit = "default"

    visit_entry_func = root["entry_funcs"].get(visit)
    if not visit_entry_func:
        sys.exit("error: unknown visit function: %s; at %s" %
                 (meta_inf["visit"], parents + [key]))

    return visit_entry_func(root, parents, data, meta, coll,
                            key, val, meta_val, meta_inf, level=level)

def retrieve_data(context, path):
    """This simple implementation for testing just pretends that metadata
       is data."""
    return retrieve_meta(context, path)

def retrieve_meta(context, path):
    """Retrieves the parsed json metadata that corresponds to
       the given parts of an ns_server URL.
       A path can look like '/pools/default?not=used&ignored=yes'."""
    with open(meta_path(context, path)) as f:
        return json.loads(f.read())

"""Functions used when visit needs to retrieve data, metadata, etc.
"""
VISIT_RETRIEVE_FUNCS = {"retrieve_data": retrieve_data,
                        "retrieve_meta": retrieve_meta}

def meta_path(context, path):
    """Calculates the path of the metadata JSON file given a data/URL path."""
    path = path.split('?')[0]
    path = re.sub("/buckets/[^/]+/", "/buckets/BUCKET/", path)
    path = re.sub("/buckets/[^/]+$", "/buckets/BUCKET", path)
    path = re.sub("/nodes/[^/]+/", "/nodes/HOST%3APORT/", path)
    path = re.sub("/nodes/[^/]+$", "/nodes/HOST%3APORT", path)
    fname = os.path.dirname(__file__) \
            + "/ns_server/2.0.0/%s.json" % (path[1:].replace("/", "_"))
    return fname

def log(*args):
    logging.debug(" ".join([str(x) for x in args]))

def debug(*args):
    return

def hyphen_last(s):
    """Converts '-foo' to 'foo-', so that 'foo-' comes after 'foo'
       which is useful for sorting."""
    if s.startswith('-'):
        return s[1:] + '-'
    return s

def store_fast(root, parents, data, meta, coll,
               key, val, meta_val, meta_inf, level):
    """Instead of this sample callback, a real implementation would store
       numeric or boolean metric into fast-changing time-series DB."""
    root["run"]["tot_fast"] += 1

def store_slow(root, parents, data, meta, coll,
               key, val, meta_val, meta_inf, level):
    """Example callback when visit encounters a slow-changing value."""
    coll[key] = val
    root["run"]["tot_slow"] += 1

def url_before(context, path):
    """Example callback invoked by visit() on a new URL."""
    log("=====", path)
    return context, path

def url_after(context, path, root):
    """Example callback invoked by visit() after visiting a URL."""
    log("tot_fast =", root["run"]["tot_fast"])
    log("-----", path)
    log(json.dumps(root["run"]["coll"], sort_keys=True, indent=4))

def visit_queue(queue):
    """Visits all the URL's on a queue, potentially concurrently."""
    while queue:
        context, path = queue.pop()
        if (context.get("ctl") or {}).get("stop"):
            return context.get("ctl").get("stop")
        context, path = context["callbacks"]["url_before"](context, path)
        root = visit_url(context, path)
        context["callbacks"]["url_after"](context, path, root)

# ----------------------------------------------------------------

def main(host, port, path, store, callbacks,
         collection_funcs=VISIT_COLLECTION_FUNCS,
         retrieve_funcs=VISIT_RETRIEVE_FUNCS,
         entry_funcs=VISIT_ENTRY_FUNCS,
         strip_meta=True, ctl=None, queue=None):
    """The ease-of-use entry-point to start a recursive URL visit()."""
    context = make_context(host, port, path, store, callbacks,
                           collection_funcs=collection_funcs,
                           retrieve_funcs=retrieve_funcs,
                           entry_funcs=entry_funcs,
                           strip_meta=strip_meta, ctl=ctl, queue=queue)
    context["queue"].append((context, path))
    return visit_queue(context["queue"])

def make_context(host, port, path, store, callbacks,
                 collection_funcs, retrieve_funcs, entry_funcs,
                 strip_meta, ctl, queue):
    """Returns a context object which is passed around by visit()."""
    return {"host": host, "port": port, "store": store,
            "path": path, "queue": queue or [], "ctl": ctl,
            "collection_funcs": collection_funcs,
            "retrieve_funcs": retrieve_funcs,
            "entry_funcs": entry_funcs,
            "strip_meta": strip_meta,
            "callbacks": callbacks}

if __name__ == '__main__':
    main("127.0.0.1", 8091, "/pools/default",
         {"fast": store_fast,
          "slow": store_slow},
         {"url_before": url_before,
          "url_after": url_after})

